import json
import os
import warnings
from collections import Counter
from pathlib import Path
from typing import Any, Callable, Iterable, List, Optional, TypeVar, Union

from loguru import logger

from edsnlp import registry
from edsnlp.core.lazy_collection import LazyCollection
from edsnlp.data.base import BaseReader, BaseWriter
from edsnlp.data.converters import (
    FILENAME,
    get_dict2doc_converter,
    get_doc2dict_converter,
)
from edsnlp.utils.collections import flatten_once
from edsnlp.utils.file_system import FileSystem, normalize_fs_path, walk_match


class JsonReader(BaseReader):
    DATA_FIELDS = ()

    def __init__(
        self,
        path: Union[str, Path],
        *,
        keep_ipynb_checkpoints: bool,
        read_in_worker: bool,
        filesystem: Optional[FileSystem] = None,
    ):
        super().__init__()

        self.fs, self.path = normalize_fs_path(filesystem, path)
        self.files = (
            [
                file
                for file in walk_match(self.fs, os.path.dirname(self.path), "*.json*")
                if keep_ipynb_checkpoints or ".ipynb_checkpoints" not in str(file)
            ]
            if self.fs.isdir(self.path)
            else [self.path]
        )
        self.keep_ipynb_checkpoints = keep_ipynb_checkpoints
        self.read_in_worker = read_in_worker
        for file in self.files:
            if not self.fs.exists(file):
                raise FileNotFoundError(f"File {file} does not exist")
        assert len(self.files), f"No .json* file found under {self.path}"
        logger.info(
            f"Found {len(self.files)} file{'s' if len(self.files) > 1 else ''} "
            f"under {self.path}"
        )

    # TODO: implement read in worker = True / False

    def read_file(self, file: str):
        with self.fs.open(file, "r", encoding="utf8") as f:
            return (
                [line for line in f]
                if os.path.splitext(file)[1].startswith(".jsonl")
                else [f.read()]
            )

    def read_main(self):
        if self.read_in_worker:
            # read in worker -> each task is a file to read from
            return (
                (f, 0 if os.path.splitext(f)[1].startswith(".jsonl") else 1)
                for f in self.files
            )
        else:
            # read in worker -> each task is a non yet parsed line
            return (
                ((f, line, os.path.splitext(f)[1].startswith(".jsonl")), 1)
                for f in self.files
                for line in self.read_file(f)
            )

    def read_worker(self, tasks):
        results = []
        for task in tasks:
            if self.read_in_worker:
                filename = task
                content = self.read_file(filename)
                is_jsonl = os.path.splitext(filename)[1].startswith(".jsonl")
            else:
                filename, content, is_jsonl = task
                content = [content]
            try:
                for line in content:
                    obj = json.loads(line)
                    if not is_jsonl:
                        obj[FILENAME] = filename
                    results.append(obj)
            except Exception:
                raise Exception(f"Cannot parse {filename}")
        return results


T = TypeVar("T")


class JsonWriter(BaseWriter):
    def __init__(
        self,
        path: Union[str, Path],
        *,
        lines: Optional[bool] = None,
        overwrite: bool = False,
        filesystem: Optional[FileSystem] = None,
    ):
        self.fs, self.path = normalize_fs_path(filesystem, path)

        path_exists = self.fs.exists(self.path)
        path_is_file = path_exists and self.fs.isfile(self.path)
        might_be_file = path_exists and Path(self.path).suffix != "" or path_is_file

        lines = lines if lines is not None else might_be_file
        self.lines = lines

        assert not path_exists or (might_be_file == lines), (
            f"To save as a single jsonl file, the path must be a file and lines "
            f"must be True. To save as a directory of json files, the path must "
            f"be a directory and lines must be False. "
            f"Got path={self.path} and lines={lines}."
        )
        if path_exists:
            if self.fs.isdir(self.path):
                files = [f for f in walk_match(self.fs, self.path, "*.json*")]
                if files:
                    if not overwrite:
                        raise FileExistsError(
                            f"Directory {self.path} already exists and appear to "
                            f"contain annotations:"
                            + "".join(
                                f"\n - {s[1:]}: {v} files"
                                for s, v in Counter(
                                    Path(f).suffix for f in files
                                ).items()
                            )
                            + "\nUse overwrite=True to write anyway."
                        )
                    for f in files:
                        self.fs.rm_file(f)

            elif not overwrite:
                raise FileExistsError(
                    f"File {self.path} already exists. Use overwrite=True to write "
                    "anyway."
                )

        if might_be_file != lines:
            warnings.warn(
                f"You set lines to {lines} but the path ({self.path}) you provided "
                f"looks like a {'file' if might_be_file else 'directory'}. "
                f"To save your documents as a single jsonl file, the path must be a "
                f"file and lines must be True. To save as a directory of json files, "
                f"the path must be a directory and lines must be False. "
            )

        super().__init__()

    def write_worker(self, records):
        # If write as jsonl, we will perform the actual writing in the `write` method
        if self.lines:
            results = []
            for rec in records:
                rec.pop(FILENAME, None)
                results.append(json.dumps(rec))
            return results, len(results)
        else:
            results = []
            for rec in records:
                filename = rec.pop(FILENAME, None)
                if filename is None:
                    raise KeyError(
                        "Cannot write to a directory of json files if the "
                        "FILENAME field is not present in the records. This is likely "
                        "caused by the `note_id` attribute (used as the filename stem) "
                        "not being set on the doc."
                    )

                file_path = os.path.join(self.path, f"{filename}.json")
                self.fs.makedirs(os.path.dirname(file_path), exist_ok=True)
                with self.fs.open(file_path, "w") as f:
                    json.dump(rec, f)
                results.append(file_path)
            return results, len(results)

    def write_main(self, fragments: Iterable[Union[List[Path], List[str]]]):
        fragments = list(flatten_once(fragments))
        if self.lines:
            with self.fs.open(self.path, "w") as f:
                f.write("\n".join(fragments))
            return [self.path]
        else:
            return [f for f in fragments]


@registry.readers.register("json")
def read_json(
    path: Union[str, Path],
    converter: Optional[Union[str, Callable]] = None,
    *,
    keep_ipynb_checkpoints: bool = False,
    read_in_worker: bool = False,
    filesystem: Optional[FileSystem] = None,
    **kwargs,
) -> LazyCollection:
    """
    The JsonReader (or `edsnlp.data.read_json`) reads a directory of JSON files and
    yields documents. At the moment, only entities and attributes are loaded.

    Example
    -------
    ```{ .python .no-check }

    import edsnlp

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(...)
    doc_iterator = edsnlp.data.read_json("path/to/json/dir", converter="omop")
    annotated_docs = nlp.pipe(doc_iterator)
    ```

    !!! note "Generator vs list"

        `edsnlp.data.read_json` returns a
        [LazyCollection][edsnlp.core.lazy_collection.LazyCollection].
        To iterate over the documents multiple times efficiently or to access them by
        index, you must convert it to a list

        ```{ .python .no-check }
        docs = list(edsnlp.data.read_json("path/to/json/dir", converter="omop")
        ```

    Parameters
    ----------
    path: Union[str, Path]
        Path to the directory containing the JSON files (will recursively look for
        files in subdirectories).
    keep_ipynb_checkpoints: bool
        Whether to keep the files have ".ipynb_checkpoints" in their path.
    read_in_worker: bool
        Whether to read the files in the worker or in the main process.
    converter: Optional[Union[str, Callable]]
        Converter to use to convert the JSON objects to Doc objects.
        These are documented on the [Converters](/data/converters) page.
    filesystem: Optional[FileSystem]
        The filesystem to use to write the files. If None, the filesystem will be
        inferred from the path (e.g. `s3://` will use S3).
    kwargs:
        Additional keyword arguments to pass to the converter. These are documented on
        the [Converters](/data/converters) page.

    Returns
    -------
    LazyCollection
    """
    data = LazyCollection(
        reader=JsonReader(
            path,
            keep_ipynb_checkpoints=keep_ipynb_checkpoints,
            read_in_worker=read_in_worker,
            filesystem=filesystem,
        )
    )
    if converter:
        converter, kwargs = get_dict2doc_converter(converter, kwargs)
        data = data.map(converter, kwargs=kwargs)
    return data


@registry.writers.register("json")
def write_json(
    data: Union[Any, LazyCollection],
    path: Union[str, Path],
    *,
    lines: bool = None,
    overwrite: bool = False,
    converter: Optional[Union[str, Callable]] = None,
    filesystem: Optional[FileSystem] = None,
    **kwargs,
) -> None:
    """
    `edsnlp.data.write_json` writes a list of documents using the JSON
    format in a directory. If `lines` is false, each document will be stored in its
    own JSON file, named after the FILENAME field returned by the converter (commonly
    the `note_id` attribute of the documents), and subdirectories will be created if the
    name contains `/` characters.

    Example
    -------
    ```{ .python .no-check }

    import edsnlp

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(...)

    doc = nlp("My document with entities")

    edsnlp.data.write_json([doc], "path/to/json/file", converter="omop", lines=True)
    # or to write a directory of JSON files, ensure that each doc has a doc._.note_id
    # attribute, since this will be used as a filename:
    edsnlp.data.write_json([doc], "path/to/json/dir", converter="omop", lines=False)
    ```

    !!! warning "Overwriting files"

        By default, `write_json` will raise an error if the directory already exists
        and contains files with `.a*` or `.txt` suffixes. This is to avoid overwriting
        existing annotations. To allow overwriting existing files, use `overwrite=True`.

    Parameters
    ----------
    data: Union[Any, LazyCollection],
        The data to write (either a list of documents or a LazyCollection).
    path: Union[str, Path]
        Path to either
        - a file if `lines` is true : this will write the documents as a JSONL file
        - a directory if `lines` is false: this will write one JSON file per document
          using the FILENAME field returned by the converter (commonly the `note_id`
          attribute of the documents) as the filename.
    lines: Optional[bool]
        Whether to write the documents as a JSONL file or as a directory of JSON files.
        By default, this is inferred from the path: if the path is a file, lines is
        assumed to be true, otherwise it is assumed to be false.
    overwrite: bool
        Whether to overwrite existing directories.
    converter: Optional[Union[str, Callable]]
        Converter to use to convert the documents to dictionary objects before writing
        them. These are documented on the [Converters](/data/converters) page.
    filesystem: Optional[FileSystem]
        The filesystem to use to write the files. If None, the filesystem will be
        inferred from the path (e.g. `s3://` will use S3).
    kwargs:
        Additional keyword arguments to pass to the converter. These are documented on
        the [Converters](/data/converters) page.
    """

    data = LazyCollection.ensure_lazy(data)
    if converter:
        converter, kwargs = get_doc2dict_converter(converter, kwargs)
        data = data.map(converter, kwargs=kwargs)

    return data.write(
        JsonWriter(
            path,
            lines=lines,
            overwrite=overwrite,
            filesystem=filesystem,
        )
    )



from __future__ import annotations

import sys
from contextlib import nullcontext
from typing import TYPE_CHECKING

from edsnlp.data.converters import set_current_tokenizer
from edsnlp.utils.batching import batchify_fns
from edsnlp.utils.collections import batchify, flatten

if TYPE_CHECKING:
    from edsnlp.core.lazy_collection import LazyCollection

doc_size_fns = {
    "words": len,
}


def apply_basic_pipes(docs, pipes):
    for name, pipe, kwargs, tok in pipes:
        with set_current_tokenizer(tok):
            if hasattr(pipe, "batch_process"):
                docs = pipe.batch_process(docs)
            else:
                docs = [pipe(doc, **kwargs) for doc in docs]
    return docs


def execute_simple_backend(
    lc: LazyCollection,
):
    """
    This is the default execution mode which batches the documents and processes each
    batch on the current process in a sequential manner.
    """
    try:
        no_grad = sys.modules["torch"].no_grad
    except (KeyError, AttributeError):
        no_grad = nullcontext
    reader = lc.reader
    writer = lc.writer
    show_progress = lc.show_progress

    split_into_batches_after = lc.split_into_batches_after
    if split_into_batches_after is None and (lc.batch_by != "docs" or lc.sort_chunks):
        split_into_batches_after = next(
            (s[0] for s in lc.pipeline if s[0] == "_ensure_doc"), None
        )
    names = [None] + [step[0] for step in lc.pipeline]
    chunk_components = lc.pipeline[: names.index(split_into_batches_after)]
    batch_components = lc.pipeline[names.index(split_into_batches_after) :]

    def process():
        bar = nullcontext()
        if show_progress:
            from tqdm import tqdm

            bar = tqdm(smoothing=0.1, mininterval=5.0)

        with bar, lc.eval():
            for docs in batchify(
                (
                    subtask
                    for task, count in reader.read_main()
                    for subtask in reader.read_worker([task])
                ),
                batch_size=lc.chunk_size,
            ):
                docs = apply_basic_pipes(docs, chunk_components)

                if lc.sort_chunks:
                    docs.sort(key=doc_size_fns.get(lc.sort_chunks, len))

                for batch in batchify_fns[lc.batch_by](docs, lc.batch_size):
                    count = len(batch)
                    with no_grad(), lc.cache():
                        batch = apply_basic_pipes(batch, batch_components)

                    if writer is not None:
                        result, count = writer.write_worker(batch)
                        if show_progress:
                            bar.update(count)
                        yield result
                    else:
                        if show_progress:
                            bar.update(count)
                        yield batch
            if writer is not None:
                result, count = writer.finalize()
                if show_progress:
                    bar.update(count)
                if count:
                    yield result

    gen = process()
    return flatten(gen) if writer is None else writer.write_main(gen)



main_pattern = dict(
    source="main",
    regex=[
        r"demence",
        r"dementiel",
        r"corps de le[vw]y",
        r"deficits?.chroniques?.cognitif",
        r"troubles?.mnesique?",
        r"troubles?.praxique",
        r"troubles?.attentionel",
        r"troubles?.degeneratif.{1,15}fonctions.{1,5}sup",
        r"maladies?.cerebrales?.degen",
        r"troubles?.neurocogn",
        r"deficits?.cognitif",
        r"(trouble|dysfonction).{1,20} cogniti",
        r"atteinte.{1,7}spheres?cogniti",
        r"syndrome.{1,10}(frontal|neuro.deg)",
        r"dysfonction.{1,25}cogni",
        r"(?<!specialisee )alzheimer",
        r"demence.{1,20}(\balz|\bpark)",
        r"binswanger",
        r"gehring",
        r"\bpick",
        r"de guam",
        r"[kc]reutzfeld.{1,5}ja[ck]ob",
        r"huntington",
        r"korsako[fv]",
        r"atrophie.{1,10}(cortico|hippocamp|cereb|lobe)",
    ],
)

acronym = dict(
    source="acronym",
    regex=[
        r"\bSLA\b",
        r"\bDFT\b",
        r"\bDFT",
        r"\bTNC\b",
    ],
    regex_attr="TEXT",
    exclude=dict(
        regex=r"\banti",  # anticorps
        window=-15,
        regex_attr="NORM",
    ),
)

charcot = dict(
    source="charcot",
    regex=[
        r"maladie.{1,10}charcot",
    ],
    exclude=dict(
        regex=[
            "pied de",
            "marie.?tooth",
        ],
        window=(-3, 3),
    ),
)


default_patterns = [
    main_pattern,
    acronym,
    charcot,
]



def test_covid(blank_nlp):
    examples = [
        ("Patient admis pour coronavirus", "coronavirus"),
        ("Patient admis pour pneumopathie à coronavirus", "pneumopathie à coronavirus"),
    ]

    blank_nlp.add_pipe("eds.covid")

    for example, text in examples:
        doc = blank_nlp(example)

        covid = doc.ents[0]
        assert covid.text == text



import contextlib
import functools
import importlib
import inspect
import os
import shutil
import warnings
from enum import Enum
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Container,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import spacy
import srsly
from confit import Config
from confit.errors import ConfitValidationError, patch_errors
from confit.utils.xjson import Reference
from spacy.language import BaseDefaults
from spacy.tokenizer import Tokenizer
from spacy.tokens import Doc
from spacy.util import get_lang_class
from spacy.vocab import Vocab, create_vocab
from typing_extensions import Literal

from ..core.registries import PIPE_META, CurriedFactory, FactoryMeta, registry
from ..utils.collections import (
    FrozenDict,
    FrozenList,
    FrozenNamespace,
    batch_compress_dict,
    decompress_dict,
)
from ..utils.typing import AsList
from .lazy_collection import LazyCollection

if TYPE_CHECKING:
    import torch

import edsnlp

EMPTY_LIST = FrozenList()


class CacheEnum(str, Enum):
    preprocess = "preprocess"
    collate = "collate"
    forward = "forward"


Pipe = TypeVar("Pipe", bound=Callable[[Doc], Doc])


class Pipeline:
    """
    New pipeline to use as a drop-in replacement for spaCy's pipeline.
    It uses PyTorch as the deep-learning backend and allows components to share
    subcomponents.

    See the documentation for more details.
    """

    def __init__(
        self,
        lang: str,
        create_tokenizer: Callable[["Pipeline"], Optional[Tokenizer]] = None,
        vocab: Union[bool, Vocab] = True,
        batch_size: Optional[int] = 128,
        vocab_config: Type[BaseDefaults] = None,
        meta: Dict[str, Any] = None,
    ):
        """
        Parameters
        ----------
        lang: str
            Language code
        create_tokenizer: Callable[['Pipeline'], Optional[Tokenizer]]
            Function that creates a tokenizer for the pipeline
        vocab: Union[bool, Vocab]
            Whether to create a new vocab or use an existing one
        batch_size: Optional[int]
            Batch size to use in the `.pipe()` method
        vocab_config: Type[BaseDefaults]
            Configuration for the vocab
        meta: Dict[str, Any]
            Meta information about the pipeline
        """
        spacy_blank_cls = get_lang_class(lang)

        self.Defaults = spacy_blank_cls.Defaults
        self.batch_size = batch_size
        if (vocab is not True) and (vocab_config is not None):
            raise ValueError(
                "You must specify either vocab or vocab_config but not both."
            )
        if vocab is True:
            vocab = create_vocab(lang, vocab_config or BaseDefaults)

        self.vocab = vocab

        if create_tokenizer is None:
            create_tokenizer = Config.resolve(
                spacy_blank_cls.default_config.merge(spacy_blank_cls.Defaults.config)[
                    "nlp"
                ]["tokenizer"]
            )

        self._tokenizer_config = Config.serialize(create_tokenizer)
        self.tokenizer = create_tokenizer(self)
        self._components: List[Tuple[str, Pipe]] = []
        self._disabled: List[str] = []
        self._path: Optional[Path] = None
        self.meta = dict(meta) if meta is not None else {}
        self.lang: str = lang
        self._cache: Optional[Dict] = None

    @property
    def pipeline(self) -> List[Tuple[str, Pipe]]:
        return FrozenList(self._components)

    @property
    def pipe_names(self) -> List[str]:
        return FrozenList([name for name, _ in self._components])

    @property
    def pipes(self):
        return FrozenNamespace({name: pipe for name, pipe in self._components})

    component_names = pipe_names

    def get_pipe(self, name: str) -> Pipe:
        """
        Get a component by its name.

        Parameters
        ----------
        name: str
            The name of the component to get.

        Returns
        -------
        Pipe
        """
        for n, pipe in self.pipeline:
            if n == name:
                return pipe
        raise ValueError(f"Pipe {repr(name)} not found in pipeline.")

    def has_pipe(self, name: str) -> bool:
        """
        Check if a component exists in the pipeline.

        Parameters
        ----------
        name: str
            The name of the component to check.

        Returns
        -------
        bool
        """
        return any(n == name for n, _ in self.pipeline)

    def create_pipe(
        self,
        factory: str,
        name: str,
        config: Dict[str, Any] = None,
    ) -> Pipe:
        """
        Create a component from a factory name.

        Parameters
        ----------
        factory: str
            The name of the factory to use
        name: str
            The name of the component
        config: Dict[str, Any]
            The config to pass to the factory

        Returns
        -------
        Pipe
        """
        try:
            curried: CurriedFactory = Config(
                {
                    "@factory": factory,
                    **(config if config is not None else {}),
                }
            ).resolve(registry=registry)
            if name is None:
                name = inspect.signature(curried.factory).parameters.get("name").default
            if name is None or name == inspect.Parameter.empty:
                name = factory
            pipe = curried.instantiate(nlp=self, path=(name,))
        except ConfitValidationError as e:
            raise e.with_traceback(None)
        return pipe

    def add_pipe(
        self,
        factory: Union[str, Pipe],
        first: bool = False,
        before: Optional[str] = None,
        after: Optional[str] = None,
        name: Optional[str] = None,
        config: Optional[Dict[str, Any]] = None,
    ) -> Pipe:
        """
        Add a component to the pipeline.

        Parameters
        ----------
        factory: Union[str, Pipe]
            The name of the component to add or the component itself
        name: Optional[str]
            The name of the component. If not provided, the name of the component
            will be used if it has one (.name), otherwise the factory name will be used.
        first: bool
            Whether to add the component to the beginning of the pipeline. This argument
            is mutually exclusive with `before` and `after`.
        before: Optional[str]
            The name of the component to add the new component before. This argument is
            mutually exclusive with `after` and `first`.
        after: Optional[str]
            The name of the component to add the new component after. This argument is
            mutually exclusive with `before` and `first`.
        config: Dict[str, Any]
            The arguments to pass to the component factory.

            Note that instead of replacing arguments with the same keys, the config
            will be merged with the default config of the component. This means that
            you can override specific nested arguments without having to specify the
            entire config.

        Returns
        -------
        Pipe
            The component that was added to the pipeline.
        """
        if isinstance(factory, str):
            pipe = self.create_pipe(factory, name, config)
            name = name or getattr(pipe, "name", factory)
        else:
            if config is not None:
                raise ValueError(
                    "Can't pass config or name with an instantiated component",
                )
            if isinstance(factory, CurriedFactory):
                name = name or factory.kwargs.get("name")
                factory = factory.instantiate(nlp=self, path=(name,))

            pipe = factory
            name = name or getattr(pipe, "name", None)
            if name is None:
                raise ValueError(
                    "The component does not have a name, so you must provide one",
                )
        if hasattr(pipe, "name") and name != pipe.name:
            pipe.name = name
        assert sum([before is not None, after is not None, first]) <= 1, (
            "You can only use one of before, after, or first",
        )
        insertion_idx = (
            0
            if first
            else self.pipe_names.index(before)
            if before is not None
            else self.pipe_names.index(after) + 1
            if after is not None
            else len(self._components)
        )
        self._components.insert(insertion_idx, (name, pipe))
        return pipe

    def get_pipe_meta(self, name: str) -> FactoryMeta:
        """
        Get the meta information for a component.

        Parameters
        ----------
        name: str
            The name of the component to get the meta for.

        Returns
        -------
        Dict[str, Any]
        """
        pipe = self.get_pipe(name)
        return PIPE_META.get(pipe, {})

    def make_doc(self, text: str) -> Doc:
        """
        Create a Doc from text.

        Parameters
        ----------
        text: str
            The text to create the Doc from.

        Returns
        -------
        Doc
        """
        return self.tokenizer(text)

    def _ensure_doc(self, text: Union[str, Doc]) -> Doc:
        """
        Ensure that the input is a Doc.

        Parameters
        ----------
        text: Union[str, Doc]
            The text to create the Doc from, or a Doc.

        Returns
        -------
        Doc
        """
        return text if isinstance(text, Doc) else self.make_doc(text)

    def __call__(self, text: Union[str, Doc]) -> Doc:
        """
        Apply each component successively on a document.

        Parameters
        ----------
        text: Union[str, Doc]
            The text to create the Doc from, or a Doc.

        Returns
        -------
        Doc
        """
        doc = self._ensure_doc(text)

        with self.cache():
            for name, pipe in self.pipeline:
                if name in self._disabled:
                    continue
                # This is a hack to get around the ambiguity
                # between the __call__ method of Pytorch modules
                # and the __call__ methods of spacy components
                if hasattr(pipe, "batch_process"):
                    doc = next(iter(pipe.batch_process([doc])))
                else:
                    doc = pipe(doc)

        return doc

    def pipe(
        self,
        inputs: Union[Iterable, LazyCollection],
        batch_size: Optional[int] = None,
        n_process: int = None,
    ) -> LazyCollection:
        """
        Process a stream of documents by applying each component successively on
        batches of documents.

        Parameters
        ----------
        inputs: Iterable[Union[str, Doc]]
            The inputs to create the Docs from, or Docs directly.
        batch_size: Optional[int]
            The batch size to use. If not provided, the batch size of the pipeline
            object will be used.
        n_process: int
            Deprecated. Use the ".set(num_cpu_workers=n_process)" method on the returned
            data lazy collection instead.
            The number of parallel workers to use. If 0, the operations will be
            executed sequentially.

        Returns
        -------
        LazyCollection
        """

        if batch_size is None:
            batch_size = self.batch_size

        lazy_collection = (
            LazyCollection.ensure_lazy(inputs)
            .map_pipeline(self)
            .set_processing(batch_size=batch_size)
        )
        if n_process is not None:
            lazy_collection = lazy_collection.set_processing(num_cpu_workers=n_process)

        return lazy_collection

    @contextlib.contextmanager
    def cache(self):
        """
        Enable caching for all (trainable) components in the pipeline
        """
        for name, pipe in self.torch_components():
            pipe.enable_cache()
        yield
        for name, pipe in self.torch_components():
            pipe.disable_cache()

    def torch_components(
        self, disable: Container[str] = ()
    ) -> Iterable[Tuple[str, "edsnlp.core.torch_component.TorchComponent"]]:
        """
        Yields components that are PyTorch modules.

        Parameters
        ----------
        disable: Container[str]
            The names of disabled components, which will be skipped.

        Returns
        -------
        Iterable[Tuple[str, 'edsnlp.core.torch_component.TorchComponent']]
        """
        for name, pipe in self.pipeline:
            if name not in disable and hasattr(pipe, "forward"):
                yield name, pipe

    def post_init(self, data: Iterable[Doc], exclude: Optional[Set] = None):
        """
        Completes the initialization of the pipeline by calling the post_init
        method of all components that have one.
        This is useful for components that need to see some data to build
        their vocabulary, for instance.

        Parameters
        ----------
        data: Iterable[Doc]
            The documents to use for initialization.
            Each component will not necessarily see all the data.
        exclude: Optional[Container]
            Components to exclude from post initialization on data
        """
        exclude = set() if exclude is None else set(exclude)
        for name, pipe in self._components:
            if hasattr(pipe, "post_init"):
                pipe.post_init(data, exclude=exclude)

    @classmethod
    def from_config(
        cls,
        config: Dict[str, Any] = {},
        *,
        vocab: Union[Vocab, bool] = True,
        disable: Union[str, Iterable[str]] = EMPTY_LIST,
        enable: Union[str, Iterable[str]] = EMPTY_LIST,
        exclude: Union[str, Iterable[str]] = EMPTY_LIST,
        meta: Dict[str, Any] = FrozenDict(),
    ):
        """
        Create a pipeline from a config object

        Parameters
        ----------
        config: Config
            The config to use
        vocab: Union["spacy.Vocab", bool]
            The spaCy vocab to use. If True, a new vocab will be created
        disable: Union[str, Iterable[str]]
            Components to disable
        enable: Union[str, Iterable[str]]
            Components to enable
        exclude: Union[str, Iterable[str]]
            Components to exclude
        meta: Dict[str, Any]
            Metadata to add to the pipeline

        Returns
        -------
        Pipeline
        """
        root_config = config.copy()

        if "nlp" in config:
            if "components" in config and "components" not in config["nlp"]:
                config["nlp"]["components"] = Reference("components")
            config = config["nlp"]

        config = dict(Config(config).resolve(root=root_config))
        components = config.pop("components", {})
        pipeline = config.pop("pipeline", ())
        tokenizer = config.pop("tokenizer", None)
        disable = (config.pop("disabled", ()), disable)

        nlp = Pipeline(
            vocab=vocab,
            create_tokenizer=tokenizer,
            meta=meta,
            **config,
        )

        # Since components are actually resolved as curried factories,
        # we need to instantiate them here
        for name, component in components.items():
            if not isinstance(component, CurriedFactory):
                raise ValueError(
                    f"Component {repr(name)} is not instantiable. Please make sure "
                    "that you didn't forget to add a '@factory' key to the component "
                    "config."
                )

        try:
            components = CurriedFactory.instantiate(components, nlp=nlp)
        except ConfitValidationError as e:
            e = ConfitValidationError(
                e.raw_errors,
                model=cls,
                name=cls.__module__ + "." + cls.__qualname__,
            )
            e.raw_errors = patch_errors(e.raw_errors, ("components",))
            raise e

        for name in pipeline:
            if name in exclude:
                continue
            if name not in components:
                raise ValueError(f"Component {repr(name)} not found in config")
            nlp.add_pipe(components[name], name=name)

        # Set of components name if it's in the disable list
        # or if it's not in the enable list and the enable list is not empty
        nlp._disabled = [
            name
            for name in nlp.pipe_names
            if name in disable or (enable and name not in enable)
        ]
        return nlp

    @property
    def disabled(self):
        """
        The names of the disabled components
        """
        return FrozenList(self._disabled)

    @classmethod
    def __get_validators__(cls):
        """
        Pydantic validators generator
        """
        yield cls.validate

    @classmethod
    def validate(cls, v, config=None):
        """
        Pydantic validator, used in the `validate_arguments` decorated functions
        """
        if isinstance(v, dict):
            return cls.from_config(v)
        if not isinstance(v, cls):
            raise ValueError("input is not a Pipeline or config dict")
        return v

    @property
    def preprocess(self):
        compressor = batch_compress_dict()

        @functools.wraps(self._preprocess)
        def preprocess(doc, supervision: bool = False, compress: bool = True):
            prep = self._preprocess(doc, supervision)
            if compress:
                prep = compressor(prep)
            return prep

        return preprocess

    def _preprocess(self, doc: Doc, supervision: bool = False):
        """
        Run the preprocessing methods of each component in the pipeline
        on a document and returns a dictionary containing the results, with the
        component names as keys.

        Parameters
        ----------
        doc: InputT
            The document to preprocess
        supervision: bool
            Whether to include supervision information in the preprocessing

        Returns
        -------
        Dict[str, Any]
        """
        prep = {}
        with self.cache():
            for name, component in self.pipeline:
                if name not in self._disabled:
                    prep_comp = (
                        component.preprocess_supervised(doc)
                        if supervision and hasattr(component, "preprocess_supervised")
                        else component.preprocess(doc)
                        if hasattr(component, "preprocess")
                        else None
                    )
                    if prep_comp is not None:
                        prep[name] = prep_comp
        return prep

    def preprocess_many(self, docs: Iterable[Doc], compress=True, supervision=True):
        """
        Runs the preprocessing methods of each component in the pipeline on
        a collection of documents and returns an iterable of dictionaries containing
        the results, with the component names as keys.

        Parameters
        ----------
        docs: Iterable[InputT]
        compress: bool
            Whether to deduplicate identical preprocessing outputs of the results
            if multiple documents share identical subcomponents. This step is required
            to enable the cache mechanism when training or running the pipeline over a
            tabular datasets such as pyarrow tables that do not store referential
            equality information.
        supervision: bool
            Whether to include supervision information in the preprocessing

        Returns
        -------
        Iterable[OutputT]
        """
        res = LazyCollection.ensure_lazy(docs)
        res = res.map(functools.partial(self.preprocess, supervision=supervision))
        if compress:
            res = res.map(batch_compress_dict())
        return res

    def collate(
        self,
        batch: Union[Iterable[Dict[str, Any]], Dict[str, Any]],
    ):
        """
        Collates a batch of preprocessed samples into a single (maybe nested)
        dictionary of tensors by calling the collate method of each component.

        Parameters
        ----------
        batch: Union[Iterable[Dict[str, Any]], Dict[str, Any]]
            The batch of preprocessed samples

        Returns
        -------
        Dict[str, Any]
            The collated batch
        """
        batch = decompress_dict(batch)
        with self.cache():
            for name, component in self.pipeline:
                if name in batch:
                    component_inputs = batch[name]
                    batch[name] = component.collate(component_inputs)
        return batch

    def parameters(self):
        """Returns an iterator over the Pytorch parameters of the components in the
        pipeline"""
        return (p for n, p in self.named_parameters())

    def named_parameters(self):
        """Returns an iterator over the Pytorch parameters of the components in the
        pipeline"""
        seen = set()
        for name, component in self.pipeline:
            if hasattr(component, "named_parameters"):
                for param_name, param in component.named_parameters():
                    if param in seen:
                        continue
                    seen.add(param)
                    yield f"{name}.{param_name}", param

    def to(self, device: Union[str, Optional["torch.device"]] = None):  # noqa F821
        """Moves the pipeline to a given device"""
        for name, component in self.torch_components():
            component.to(device)
        return self

    def train(self, mode=True):
        """
        Enables training mode on pytorch modules

        Parameters
        ----------
        mode: bool
            Whether to enable training or not
        """

        class context:
            def __enter__(self):
                pass

            def __exit__(ctx_self, type, value, traceback):
                for name, proc in self.torch_components():
                    proc.train(was_training[name])

        was_training = {name: proc.training for name, proc in self.torch_components()}
        for name, proc in self.torch_components():
            proc.train(mode)

        return context()

    def to_disk(
        self, path: Union[str, Path], *, exclude: Optional[Set[str]] = None
    ) -> None:
        """
        Save the pipeline to a directory.

        Parameters
        ----------
        path: Union[str, Path]
            The path to the directory to save the pipeline to. Every component will be
            saved to separated subdirectories of this directory, except for tensors
            that will be saved to a shared files depending on the references between
            the components.
        exclude: Sequence[str]
            The names of the components, or attributes to exclude from the saving
            process.
        """
        exclude = set() if exclude is None else exclude

        path = Path(path) if isinstance(path, str) else path

        if (
            os.path.exists(path)
            and os.listdir(path)
            and not os.path.exists(path / "config.cfg")
        ):
            raise Exception(
                "The directory already exists and doesn't appear to be a"
                "saved pipeline (missing config.cfg). Please erase it manually or "
                "choose a different directory."
            )
        shutil.rmtree(path, ignore_errors=True)
        os.makedirs(path, exist_ok=True)

        if "tokenizer" not in exclude:
            self.tokenizer.to_disk(path / "tokenizer", exclude=["vocab"])
        if "meta" not in exclude:
            srsly.write_json(path / "meta.json", self.meta)
        if "vocab" not in exclude:
            self.vocab.to_disk(path / "vocab")

        pwd = os.getcwd()
        overrides = {"components": {}}
        try:
            os.chdir(path)
            for pipe_name, pipe in self._components:
                if hasattr(pipe, "to_disk") and pipe_name not in exclude:
                    pipe_overrides = pipe.to_disk(Path(pipe_name), exclude=exclude)
                    overrides["components"][pipe_name] = pipe_overrides
        finally:
            os.chdir(pwd)

        config = self.config.merge(overrides)

        if "config" not in exclude:
            config.to_disk(path / "config.cfg")

    def from_disk(
        self,
        path: Union[str, Path],
        *,
        exclude: Optional[Union[str, Sequence[str]]] = None,
        device: Optional[Union[str, "torch.device"]] = "cpu",  # noqa F821
    ) -> "Pipeline":
        """
        Load the pipeline from a directory. Components will be updated in-place.

        Parameters
        ----------
        path: Union[str, Path]
            The path to the directory to load the pipeline from
        exclude: Optional[Union[str, Sequence[str]]]
            The names of the components, or attributes to exclude from the loading
            process.
        device: Optional[Union[str, "torch.device"]]
            Device to use when loading the tensors
        """

        def deserialize_meta(path: Path) -> None:
            if path.exists():
                data = srsly.read_json(path)
                self.meta.update(data)
                # self.meta always overrides meta["vectors"] with the metadata
                # from self.vocab.vectors, so set the name directly
                self.vocab.vectors.name = data.get("vectors", {}).get("name")

        def deserialize_vocab(path: Path) -> None:
            if path.exists():
                self.vocab.from_disk(path, exclude=exclude)

        exclude = (
            set()
            if exclude is None
            else {exclude}
            if isinstance(exclude, str)
            else set(exclude)
        )

        path = (Path(path) if isinstance(path, str) else path).absolute()
        if (path / "meta.json").exists() and "meta" not in exclude:
            deserialize_meta(path / "meta.json")
        if (path / "vocab").exists() and "vocab" not in exclude:
            deserialize_vocab(path / "vocab")
        if (path / "tokenizer").exists() and "tokenizer" not in exclude:
            self.tokenizer.from_disk(path / "tokenizer", exclude=["vocab"])

        pwd = os.getcwd()
        try:
            os.chdir(path)
            for name, proc in self._components:
                if hasattr(proc, "from_disk") and name not in exclude:
                    proc.from_disk(Path(name), exclude=exclude)
                # Convert to list here in case exclude is (default) tuple
                exclude.add(name)
        finally:
            os.chdir(pwd)

        self._path = path  # type: ignore[assignment]
        return self

    # override config property getter to remove "factory" key from components
    @property
    def cfg(self) -> Config:
        """
        Returns the config of the pipeline, including the config of all components.
        Updated from spacy to allow references between components.
        """
        return Config(
            {
                "lang": self.lang,
                "pipeline": list(self.pipe_names),
                "components": {key: component for key, component in self._components},
                "tokenizer": self._tokenizer_config,
            }
        )

    @property
    def config(self) -> Config:
        config = Config({"nlp": self.cfg.copy()})
        config["components"] = config["nlp"].pop("components")
        return config.serialize()

    def select_pipes(
        self,
        *,
        disable: Optional[Union[str, Iterable[str]]] = None,
        enable: Optional[Union[str, Iterable[str]]] = None,
    ):
        """
        Temporarily disable and enable components in the pipeline.

        Parameters
        ----------
        disable: Optional[Union[str, Iterable[str]]]
            The name of the component to disable, or a list of names.
        enable: Optional[Union[str, Iterable[str]]]
            The name of the component to enable, or a list of names.
        """

        class context:
            def __enter__(self):
                pass

            def __exit__(ctx_self, type, value, traceback):
                self._disabled = disabled_before

        disable = [disable] if isinstance(disable, str) else disable
        pipe_names = set(self.pipe_names)
        if enable is not None:
            enable = [enable] if isinstance(enable, str) else enable
            if set(enable) - pipe_names:
                raise ValueError(
                    "Enabled pipes {} not found in pipeline.".format(
                        sorted(set(enable) - pipe_names)
                    )
                )
            to_disable = [pipe for pipe in self.pipe_names if pipe not in enable]
            # raise an error if the enable and disable keywords are not consistent
            if disable is not None and disable != to_disable:
                raise ValueError("Inconsistent values for `enable` and `disable`")
            disable = to_disable

        if set(disable) - pipe_names:
            raise ValueError(
                "Disabled pipes {} not found in pipeline.".format(
                    sorted(set(disable) - pipe_names)
                )
            )

        disabled_before = self._disabled
        self._disabled = disable
        return context()

    def package(
        self,
        name: Optional[str] = None,
        root_dir: Union[str, Path] = ".",
        build_dir: Union[str, Path] = "build",
        dist_dir: Union[str, Path] = "dist",
        artifacts_name: str = "artifacts",
        check_dependencies: bool = False,
        project_type: Optional[Literal["poetry", "setuptools"]] = None,
        version: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = {},
        distributions: Optional[AsList[Literal["wheel", "sdist"]]] = ["wheel"],
        config_settings: Optional[Mapping[str, Union[str, Sequence[str]]]] = None,
        isolation: bool = True,
        skip_build_dependency_check: bool = False,
    ):
        from edsnlp.utils.package import package

        return package(
            pipeline=self,
            name=name,
            root_dir=root_dir,
            build_dir=build_dir,
            dist_dir=dist_dir,
            artifacts_name=artifacts_name,
            check_dependencies=check_dependencies,
            project_type=project_type,
            version=version,
            metadata=metadata,
            distributions=distributions,
            config_settings=config_settings,
            isolation=isolation,
            skip_build_dependency_check=skip_build_dependency_check,
        )

    def __getstate__(self):
        state = self.__dict__.copy()
        state["_pipe_meta"] = [PIPE_META.get(pipe, {}) for _, pipe in self.pipeline]
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        for (name, pipe), meta in zip(self.pipeline, state["_pipe_meta"]):
            PIPE_META[pipe] = meta
        del state["_pipe_meta"]


def blank(
    lang: str,
    config: Union[Dict[str, Any], Config] = {},
):
    """
    Loads an empty EDS-NLP Pipeline, similarly to `spacy.blank`. In addition to
    standard components, this pipeline supports EDS-NLP trainable torch components.

    Examples
    --------
    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.covid())
    ```

    Parameters
    ----------
    lang: str
        Language ID, e.g. "en", "fr", "eds", etc.
    config: Union[Dict[str, Any], Config]
        The config to use for the pipeline

    Returns
    -------
    Pipeline
        The new empty pipeline instance.
    """
    # Check if language is registered / entry point is available
    config_lang = (
        config["nlp"].get("lang")
        if isinstance(config.get("nlp"), dict)
        else config.get("lang")
    )
    if config_lang is not None and config_lang != lang:
        raise ValueError(
            "The language specified in the config does not match the lang argument."
        )
    return Pipeline.from_config({"lang": lang, **config})


def load(
    model: Union[Path, str, Config],
    overrides: Optional[Dict[str, Any]] = None,
    *,
    exclude: Optional[Union[str, Iterable[str]]] = None,
):
    """
    Load a pipeline from a config file or a directory.

    Examples
    --------

    ```{ .python .no-check }
    import edsnlp

    nlp = edsnlp.load(
        "path/to/config.cfg",
        overrides={"components": {"my_component": {"arg": "value"}}},
    )
    ```

    Parameters
    ----------
    model: Union[Path, str, Config]
        The config to use for the pipeline, or the path to a config file or a directory.
    overrides: Optional[Dict[str, Any]]
        Overrides to apply to the config when loading the pipeline. These are the
        same parameters as the ones used when initializing the pipeline.
    exclude: Optional[Union[str, Iterable[str]]]
        The names of the components, or attributes to exclude from the loading
        process. :warning: The `exclude` argument will be mutated in place.

    Returns
    -------
    Pipeline
    """
    error = (
        "The load function expects either :\n"
        "- a confit Config object\n"
        "- the path of a config file (.cfg file)\n"
        "- the path of a trained model\n"
        "- the name of an installed pipeline package\n"
        f"but got {model!r} which is neither"
    )
    if isinstance(model, (Path, str)):
        path = Path(model)
        is_dir = path.is_dir()
        is_config = path.is_file() and path.suffix == ".cfg"
        try:
            module = importlib.import_module(model)
            is_package = True
        except (ImportError, AttributeError, TypeError):
            module = None
            is_package = False
        if is_dir and is_package:
            warnings.warn(
                "The path provided is both a directory and a package : edsnlp will "
                "load the package. To load from the directory instead, please pass the "
                f'path as "./{path}" instead.'
            )
        if is_dir:
            path = (Path(path) if isinstance(path, str) else path).absolute()
            config = Config.from_disk(path / "config.cfg")
            if overrides:
                config = config.merge(overrides)
            pwd = os.getcwd()
            try:
                os.chdir(path)
                nlp = Pipeline.from_config(config)
                nlp.from_disk(path, exclude=exclude)
                nlp.train(False)
            finally:
                os.chdir(pwd)
            return nlp
        elif is_config:
            model = Config.from_disk(path)
        elif is_package:
            # Load as package
            available_kwargs = {
                "overrides": overrides,
                "exclude": exclude,
            }
            signature_kwargs = inspect.signature(module.load).parameters
            kwargs = {
                name: available_kwargs[name]
                for name in signature_kwargs
                if name in available_kwargs
            }
            return module.load(**kwargs)

    if not isinstance(model, Config):
        raise ValueError(error)

    return Pipeline.from_config(model)


PipelineProtocol = Union[Pipeline, spacy.Language]



results_cerebrovascular_accident = dict(
    has_match=[
        False,
        True,
        True,
        False,
        True,
        True,
        True,
    ],
    detailled_status="PRESENT",
    assign=None,
    texts=[
        "Patient hospitalisé à AVC.",
        "Hospitalisation pour un AVC.",
        "Saignement intracranien",
        "Thrombose périphérique",
        "Thrombose sylvienne",
        "Infarctus cérébral",
        "Soigné via un thrombolyse",
    ],
)



from typing import TYPE_CHECKING

from edsnlp.utils.lazy_module import lazify

lazify()

if TYPE_CHECKING:
    from .deprecated_pipe import pipe  # DEPRECATED
    from .spark import execute_spark_backend
    from .simple import execute_simple_backend
    from .multiprocessing import execute_multiprocessing_backend



import pickle
from typing import List, Optional, Union

import numpy as np
import pandas as pd
from spacy.tokens import Doc, Span, Token

from edsnlp.core import PipelineProtocol
from edsnlp.pipes.core.matcher.matcher import GenericMatcher
from edsnlp.utils.filter import get_spans

from .functional import build_path
from .model import EndLinesModel


class EndLinesMatcher(GenericMatcher):
    '''
    The `eds.endlines` component classifies newline characters as actual end of lines
    or mere spaces. In the latter case, the token is removed from the normalised
    document.

    Behind the scenes, it uses a `endlinesmodel` instance, which is an unsupervised
    algorithm based on the work of [@zweigenbaum2016].

    Training
    --------
    ```python
    import edsnlp
    from edsnlp.pipes.core.endlines.model import EndLinesModel

    nlp = edsnlp.blank("eds")

    texts = [
        """
    Le patient est arrivé hier soir.
    Il est accompagné par son fils

    ANTECEDENTS
    Il a fait une TS en 2010
    Fumeur, il est arreté il a 5 mois
    Chirurgie de coeur en 2011
    CONCLUSION
    Il doit prendre
    le medicament indiqué 3 fois par jour. Revoir médecin
    dans 1 mois.
    DIAGNOSTIC :

    Antecedents Familiaux:
    - 1. Père avec diabete
    """,
        """
    J'aime le
    fromage...
    """,
    ]

    docs = list(nlp.pipe(texts))

    # Train and predict an EndLinesModel
    endlines = EndLinesModel(nlp=nlp)

    df = endlines.fit_and_predict(docs)
    df.head()

    PATH = "/tmp/path_to_save"
    endlines.save(PATH)
    ```

    Examples
    --------
    ```python
    import edsnlp, edsnlp.pipes as eds
    from spacy.tokens import Span
    from spacy import displacy

    nlp = edsnlp.blank("eds")

    PATH = "/tmp/path_to_save"
    nlp.add_pipe(eds.endlines(model_path=PATH))

    docs = list(nlp.pipe(texts))

    doc_exemple = docs[1]

    doc_exemple.ents = tuple(
        Span(doc_exemple, token.i, token.i + 1, "excluded")
        for token in doc_exemple
        if token.tag_ == "EXCLUDED"
    )

    displacy.render(doc_exemple, style="ent", options={"colors": {"space": "red"}})
    ```

    Extensions
    ----------
    The `eds.endlines` pipeline declares one extension, on both `Span` and `Token`
    objects. The `end_line` attribute is a boolean, set to `True` if the pipeline
    predicts that the new line is an end line character. Otherwise, it is set to
    `False` if the new line is classified as a space.

    The pipeline also sets the `excluded` custom attribute on newlines that are
    classified as spaces. It lets downstream matchers skip excluded tokens
    (see [normalisation](/pipes/core/normalisation/)) for more detail.

    Parameters
    ----------
    nlp : PipelineProtocol
        The pipeline object.
    name: str
        The name of the component.
    model_path : Optional[Union[str, EndLinesModel]]
        Path to trained model. If None, it will use a default model

    Authors and citation
    --------------------
    The `eds.endlines` pipeline was developed by AP-HP's Data Science team based on
    the work of [@zweigenbaum2016].
    '''

    def __init__(
        self,
        nlp: PipelineProtocol,
        name: Optional[str] = "endlines",
        *,
        model_path: Optional[Union[str, EndLinesModel]] = None,
    ):
        super().__init__(
            nlp=nlp,
            name=name,
            terms=None,
            attr="TEXT",
            regex=dict(
                new_line=r"\n+",
            ),
            ignore_excluded=False,
            ignore_space_tokens=False,
        )

        self._read_model(model_path)

    def _read_model(self, end_lines_model: Optional[Union[str, EndLinesModel]]):
        """
        Parameters
        ----------
        end_lines_model : Optional[Union[str, EndLinesModel]]

        Raises
        ------
        TypeError
        """
        if end_lines_model is None:
            path = build_path(__file__, "base_model.pkl")

            with open(path, "rb") as inp:
                self.model = pickle.load(inp)
        elif isinstance(end_lines_model, str):
            with open(end_lines_model, "rb") as inp:
                self.model = pickle.load(inp)
        elif isinstance(end_lines_model, EndLinesModel):
            self.model = end_lines_model
        else:
            raise TypeError(
                "type(`end_lines_model`) should be one of {None, str, EndLinesModel}"
            )

    @classmethod
    def _spacy_compute_a3a4(cls, token: Token) -> str:
        """Function to compute A3 and A4

        Parameters
        ----------
        token : Token

        Returns
        -------
        str
        """

        if token.is_upper:
            return "UPPER"

        elif token.shape_.startswith("Xx"):
            return "S_UPPER"

        elif token.shape_.startswith("x"):
            return "LOWER"

        elif (token.is_digit) & (
            (token.doc[max(token.i - 1, 0)].is_punct)
            | (token.doc[min(token.i + 1, len(token.doc) - 1)].is_punct)
        ):
            return "ENUMERATION"

        elif token.is_digit:
            return "DIGIT"

        elif (token.is_punct) & (token.text in [".", ";", "..", "..."]):
            return "STRONG_PUNCT"

        elif (token.is_punct) & (token.text not in [".", ";", "..", "..."]):
            return "SOFT_PUNCT"

        else:
            return "OTHER"

    @classmethod
    def _compute_length(cls, doc: Doc, start: int, end: int) -> int:
        """Compute length without spaces

        Parameters
        ----------
        doc : Doc
        start : int
        end : int

        Returns
        -------
        int
        """
        length = 0
        for t in doc[start:end]:
            length += len(t.text)

        return length

    def _get_df(self, doc: Doc, new_lines: List[Span]) -> pd.DataFrame:
        """Get a pandas DataFrame to call the classifier

        Parameters
        ----------
        doc : Doc
        new_lines : List[Span]

        Returns
        -------
        pd.DataFrame
        """

        data = []
        for i, span in enumerate(new_lines):
            start = span.start
            end = span.end

            max_index = len(doc) - 1
            a1_token = doc[max(start - 1, 0)]
            a2_token = doc[min(start + 1, max_index)]
            a1 = a1_token.orth
            a2 = a2_token.orth
            a3 = self._spacy_compute_a3a4(a1_token)
            a4 = self._spacy_compute_a3a4(a2_token)
            blank_line = "\n\n" in span.text

            if i > 0:
                start_previous = new_lines[i - 1].start + 1
            else:
                start_previous = 0

            length = self._compute_length(
                doc, start=start_previous, end=start
            )  # It's ok cause i count the total length from the previous up to this one

            data_dict = dict(
                span_start=start,
                span_end=end,
                A1=a1,
                A2=a2,
                A3=a3,
                A4=a4,
                BLANK_LINE=blank_line,
                length=length,
            )
            data.append(data_dict)

        df = pd.DataFrame(data)

        mu = df["length"].mean()
        sigma = df["length"].std()
        if np.isnan(sigma):
            sigma = 1

        cv = sigma / mu
        df["B1"] = (df["length"] - mu) / sigma
        df["B2"] = cv

        return df

    def __call__(self, doc: Doc) -> Doc:
        """
        Predict for each new line if it's an end of line or a space.

        Parameters
        ----------
        doc: spaCy Doc object

        Returns
        -------
        doc: spaCy Doc object, with each new line annotated
        """

        matches = self.process(doc)
        new_lines = get_spans(matches, "new_line")

        if len(new_lines) > 0:
            df = self._get_df(doc=doc, new_lines=new_lines)
            df = self.model.predict(df)

            for span, prediction in zip(new_lines, df.PREDICTED_END_LINE):
                for t in span:
                    t.tag_ = "ENDLINE" if prediction else "EXCLUDED"
                    if prediction:
                        t._.excluded = True

        return doc



import pytest
from spacy.tokens import Span

import edsnlp
from edsnlp.scorers.ner import create_ner_exact_scorer, create_ner_token_scorer


@pytest.fixture(scope="session")
def gold_and_pred():
    nlp = edsnlp.blank("eds")

    gold_doc1 = nlp.make_doc("Le patient a le covid 19.")
    gold_doc1.ents = [Span(gold_doc1, 4, 6, label="covid")]
    gold_doc2 = nlp.make_doc("Corona: positif. Le cvid est une maladie.")
    gold_doc2.ents = [
        Span(gold_doc2, 0, 1, label="covid"),
        Span(gold_doc2, 5, 6, label="covid"),
    ]

    pred_doc1 = nlp.make_doc("Le patient a le covid 19.")
    pred_doc1.ents = [Span(pred_doc1, 4, 6, label="covid")]
    pred_doc2 = nlp.make_doc("Corona: positif. Le cvid est une maladie.")
    pred_doc2.ents = [Span(pred_doc2, 0, 2, label="covid")]

    return [gold_doc1, gold_doc2], [pred_doc1, pred_doc2]


def test_exact_ner_scorer(gold_and_pred):
    scorer = create_ner_exact_scorer("ents")
    ner_exact_score = scorer(*gold_and_pred)
    assert ner_exact_score["micro"] == {
        "p": 0.5,
        "r": 1 / 3,
        "f": 0.4,
        "support": 3,
        "positives": 2,
        "tp": 1,
    }


def test_token_ner_scorer(gold_and_pred):
    scorer = create_ner_token_scorer("ents")
    ner_exact_score = scorer(*gold_and_pred)
    assert ner_exact_score["micro"] == {
        "f": 0.75,
        "p": 0.75,
        "r": 0.75,
        "support": 4,
        "positives": 4,
        "tp": 3,
    }



from typing import Dict, Iterable, List, Optional, Union

from loguru import logger
from spacy.tokens import Doc, Span

from edsnlp.core import PipelineProtocol
from edsnlp.pipes.core.matcher.matcher import GenericMatcher
from edsnlp.pipes.misc.reason import patterns
from edsnlp.utils.filter import get_spans
from edsnlp.utils.inclusion import check_inclusion


class ReasonMatcher(GenericMatcher):
    '''
    The `eds.reason` matcher uses a rule-based algorithm to detect spans that relate
    to the reason of the hospitalisation. It was designed at AP-HP's EDS.

    Examples
    --------
    The following snippet matches a simple terminology, and looks for spans of
    hospitalisation reasons. It is complete and can be run _as is_.

    ```python
    import edsnlp, edsnlp.pipes as eds

    text = """COMPTE RENDU D'HOSPITALISATION du 11/07/2018 au 12/07/2018
    MOTIF D'HOSPITALISATION
    Monsieur Dupont Jean Michel, de sexe masculin, âgée de 39 ans, née le 23/11/1978,
    a été hospitalisé du 11/08/2019 au 17/08/2019 pour attaque d'asthme.

    ANTÉCÉDENTS
    Antécédents médicaux :
    Premier épisode d'asthme en mai 2018."""

    nlp = edsnlp.blank("eds")

    # Extraction of entities
    nlp.add_pipe(
        eds.matcher(
            terms=dict(
                respiratoire=[
                    "asthmatique",
                    "asthme",
                    "toux",
                ]
            )
        ),
    )
    nlp.add_pipe(eds.normalizer())
    nlp.add_pipe(eds.reason(use_sections=True))
    doc = nlp(text)

    reason = doc.spans["reasons"][0]
    reason
    # Out: hospitalisé du 11/08/2019 au 17/08/2019 pour attaque d'asthme.

    reason._.is_reason
    # Out: True

    entities = reason._.ents_reason
    entities
    # Out: [asthme]

    entities[0].label_
    # Out: 'respiratoire'

    ent = entities[0]
    ent._.is_reason
    # Out: True
    ```

    Extensions
    ----------
    The `eds.reason` pipeline adds the key `reasons` to `doc.spans` and declares
    one extension, on the `Span` objects called `ents_reason`.

    The `ents_reason` extension is a list of named entities that overlap the `Span`,
    typically entities found in upstream components like `matcher`.

    It also declares the boolean extension `is_reason`. This extension is set to True
    for the Reason Spans but also for the entities that overlap the reason span.

    Parameters
    ----------
    nlp : PipelineProtocol
        The pipeline object
    name : str
        Name of the component
    reasons : Dict[str, Union[List[str], str]]
        Reason patterns
    attr : str
        Default token attribute to use to build the text to match on.
    use_sections : bool,
        Whether or not use the `sections` matcher to improve results.
    ignore_excluded : bool
        Whether to skip excluded tokens.

    Authors and citation
    --------------------
    The `eds.reason` matcher was developed by AP-HP's Data Science team.

    '''

    def __init__(
        self,
        nlp: PipelineProtocol,
        name: str = "reason",
        *,
        reasons: Optional[Dict[str, Union[List[str], str]]] = None,
        attr: Union[Dict[str, str], str] = "TEXT",
        use_sections: bool = False,
        ignore_excluded: bool = False,
    ):
        if reasons is None:
            reasons = patterns.reasons

        super().__init__(
            nlp,
            name=name,
            terms=None,
            regex=reasons,
            attr=attr,
            ignore_excluded=ignore_excluded,
            span_setter={},
        )

        self.use_sections = use_sections and (
            "eds.sections" in nlp.pipe_names or "sections" in nlp.pipe_names
        )
        if use_sections and not self.use_sections:
            logger.warning(
                "You have requested that the pipeline use annotations "
                "provided by the `eds.section` pipeline, but it was not set. "
                "Skipping that step."
            )

        self.set_extensions()

    @classmethod
    def set_extensions(cls) -> None:
        if not Span.has_extension("ents_reason"):
            Span.set_extension("ents_reason", default=None)

        # TODO: remove this extension, and filter directly in span group
        if not Span.has_extension("is_reason"):
            Span.set_extension("is_reason", default=False)

    def _enhance_with_sections(self, sections: Iterable, reasons: Iterable) -> List:
        """Enhance the list of reasons with the section information.
        If the reason overlaps with history, so it will be removed from the list

        Parameters
        ----------
        sections : Iterable
            Spans of sections identified with the `sections` pipeline
        reasons : Iterable
            Reasons list identified by the regex

        Returns
        -------
        List
            Updated list of spans reasons
        """

        for section in sections:
            if section.label_ in patterns.sections_reason:
                reasons.append(section)

            if section.label_ in patterns.section_exclude:
                for reason in reasons:
                    if check_inclusion(reason, section.start, section.end):
                        reasons.remove(reason)

        return reasons

    def __call__(self, doc: Doc) -> Doc:
        """Find spans related to the reasons of the hospitalisation

        Parameters
        ----------
        doc : Doc

        Returns
        -------
        Doc
        """
        matches = self.process(doc)
        reasons = get_spans(matches, "reasons")

        if self.use_sections:
            sections = doc.spans["sections"]
            reasons = self._enhance_with_sections(sections=sections, reasons=reasons)

        doc.spans["reasons"] = reasons

        # Entities
        if len(doc.ents) > 0:
            for reason in reasons:  # TODO optimize this iteration
                ent_list = []
                for ent in doc.ents:
                    if check_inclusion(ent, reason.start, reason.end):
                        ent_list.append(ent)
                        ent._.is_reason = True

                reason._.ents_reason = ent_list
                reason._.is_reason = True

        return doc



from edsnlp.core import registry

from .connective_tissue_disease import ConnectiveTissueDiseaseMatcher
from .patterns import default_patterns

DEFAULT_CONFIG = dict(
    patterns=default_patterns,
    label="connective_tissue_disease",
    span_setter={"ents": True, "connective_tissue_disease": True},
)

create_component = registry.factory.register(
    "eds.connective_tissue_disease",
    assigns=["doc.ents", "doc.spans"],
)(ConnectiveTissueDiseaseMatcher)



results_myocardial_infarction = dict(
    has_match=[
        True,
        False,
        True,
        False,
        True,
    ],
    detailled_status="PRESENT",
    assign=None,
    texts=[
        "Une cardiopathie ischémique",
        "Une cardiopathie non-ischémique",
        "Présence d'un stent sur la marginale",
        "Présence d'un stent périphérique",
        "infarctus du myocarde",
    ],
)



mild = dict(
    source="mild",
    regex=[
        r"cholangites?.{1,10}(sclero|secondaire)",
        r"fibrose.{1,10}(hepatique|foie)",
        r"hepatite.{1,15}chronique",
        r"hepatopathie",
        r"\bnash\b",
        r"(maladie|sydrome).{1,10}Hanot",
        r"surinfections.{1,5}delta",
        r"\bcbp\b",
        r"\bmaf\b",
        r"(maladie|syndrome).{1,8}hanot",
    ],
    regex_attr="NORM",
    exclude=dict(
        regex="\bdots?\b",
        window=-5,
    ),
)

moderate_severe = dict(
    source="moderate_severe",
    regex=[
        r"cirrhose",
        r"necrose.{1,10}(hepati|foie)",
        r"varice.{1,10}(estomac|oesopha|gastr)",
        r"\bvo\b.{1,5}(stade|grade).(1|2|3|i{1,3})",
        r"hypertension.{1,5}portale",
        r"scleroses.{1,5}hepatoportale",
        r"sydrome.{1,10}hepato.?ren",
        r"insuffisance.{1,5}hepa",
        r"encephalopathie.{1,5}hepa",
        r"\btips\b",
    ],
    regex_attr="NORM",
)

transplant = dict(
    source="transplant",
    regex=[
        r"(?<!pre.?)(greffe|transplant).{1,12}(hepatique|foie)",
    ],
    regex_attr="NORM",
    exclude=dict(
        regex="chc",
        window=(-5, 5),
    ),
)

default_patterns = [
    mild,
    moderate_severe,
    transplant,
]



from collections import defaultdict
from typing import Any, Dict, Iterable

from spacy.training import Example

from edsnlp import registry
from edsnlp.scorers import average_precision, make_examples, prf
from edsnlp.utils.bindings import BINDING_GETTERS, Qualifiers, QualifiersArg
from edsnlp.utils.span_getters import SpanGetterArg, get_spans


def span_classification_scorer(
    examples: Iterable[Example],
    span_getter: SpanGetterArg,
    qualifiers: Qualifiers,
    include_falsy: bool = False,
    default_values: Dict = {},
    micro_key: str = "micro",
):
    """
    Scores the extracted entities that may be overlapping or nested
    by looking in `doc.ents`, and `doc.spans`.

    Parameters
    ----------
    examples : Iterable[Example]
        The examples to score
    span_getter : SpanGetterArg
        The span getter to use to extract the spans from the document
    qualifiers : Sequence[str]
        The qualifiers to use to score the spans
    default_values: Dict
        Values to dismiss when computing the micro-average per label. This is
        useful to compute precision and recall for certain attributes that have
        imbalanced value repartitions, such as "negation", "family related"
        or "certainty" attributes.
    include_falsy : bool
        Whether to count predicted or gold occurrences of falsy values when computing
        the metrics. If `False`, only the non-falsy values will be counted and matched
        together.
    micro_key : str
        The key to use to store the micro-averaged results for spans of all types

    Returns
    -------
    Dict[str, float]
    """
    labels = defaultdict(lambda: (set(), set(), dict()))
    labels["micro"] = (set(), set(), dict())
    total_pred_count = 0
    total_gold_count = 0

    if not include_falsy:
        default_values_ = defaultdict(lambda: False)
        default_values_.update(default_values)
        default_values = default_values_
        del default_values_
    for eg_idx, eg in enumerate(examples):
        doc_spans = get_spans(eg.predicted, span_getter)
        for span_idx, span in enumerate(doc_spans):
            total_pred_count += 1
            for qualifier, span_filter in qualifiers.items():
                if not (span_filter is True or span.label_ in span_filter):
                    continue
                getter_key = (
                    qualifier if qualifier.startswith("_.") else f"_.{qualifier}"
                )
                value = BINDING_GETTERS[getter_key](span)
                top_val, top_p = max(
                    getattr(span._, "prob", {}).get(qualifier, {}).items(),
                    key=lambda x: x[1],
                    default=(value, 1.0),
                )
                if (top_val or include_falsy) and default_values[qualifier] != top_val:
                    labels[qualifier][2][(eg_idx, span_idx, qualifier, top_val)] = top_p
                    labels[micro_key][2][(eg_idx, span_idx, qualifier, top_val)] = top_p
                if (value or include_falsy) and default_values[qualifier] != value:
                    labels[micro_key][0].add((eg_idx, span_idx, qualifier, value))
                    labels[qualifier][0].add((eg_idx, span_idx, qualifier, value))

        doc_spans = get_spans(eg.reference, span_getter)
        for span_idx, span in enumerate(doc_spans):
            total_gold_count += 1
            for qualifier, span_filter in qualifiers.items():
                if not (span_filter is True or span.label_ in span_filter):
                    continue
                getter_key = (
                    qualifier if qualifier.startswith("_.") else f"_.{qualifier}"
                )
                value = BINDING_GETTERS[getter_key](span)
                if (value or include_falsy) and default_values[qualifier] != value:
                    labels[micro_key][1].add((eg_idx, span_idx, qualifier, value))
                    labels[qualifier][1].add((eg_idx, span_idx, qualifier, value))

    if total_pred_count != total_gold_count:
        raise ValueError(
            f"Number of predicted and gold spans differ: {total_pred_count} != "
            f"{total_gold_count}. Make sure that you are running your span "
            "qualifier pipe on the gold annotations, and not spans predicted by "
            "another NER pipe in your model."
        )

    return {
        name: {
            **prf(pred, gold),
            "ap": average_precision(pred_with_prob, gold),
        }
        for name, (pred, gold, pred_with_prob) in labels.items()
    }


@registry.scorers.register("eds.span_classification_scorer")
class create_span_classification_scorer:
    qualifiers: Qualifiers

    def __init__(
        self,
        span_getter: SpanGetterArg,
        qualifiers: QualifiersArg = None,
        default_values: Dict = {},
        include_falsy: bool = False,
        micro_key: str = "micro",
    ):
        self.span_getter = span_getter
        self.qualifiers = qualifiers
        self.default_values = default_values
        self.include_falsy = include_falsy
        self.micro_key = micro_key

    def __call__(self, *examples: Any):
        return span_classification_scorer(
            make_examples(*examples),
            span_getter=self.span_getter,
            qualifiers=self.qualifiers,
            default_values=self.default_values,
            include_falsy=self.include_falsy,
            micro_key=self.micro_key,
        )



results_connective_tissue_disease = dict(
    has_match=[
        True,
        True,
        False,
        True,
        True,
    ],
    detailled_status="PRESENT",
    assign=None,
    texts=[
        "Présence d'une sclérodermie.",
        "Patient atteint d'un lupus.",
        "Présence d'anticoagulants lupiques,",
        "Il y a une MICI.",
        "Syndrome de Raynaud",
    ],
)



from typing import Optional

import torch
import torch.nn.functional as F
from typing_extensions import Literal


class Metric(torch.nn.Module):
    """
    Metric layer, used for computing similarities between two sets of vectors. A typical
    use case is to compute the similarity between a set of query vectors (input
    embeddings) and a set of concept vectors (output embeddings).

    Parameters
    ----------
    in_features : int
        Size of the input embeddings
    out_features : int
        Size of the output embeddings
    num_groups : int
        Number of groups for the output embeddings, that can be used to filter out
        certain concepts that are not relevant for a given query (e.g. do not compare
        a drug with concepts for diseases)
    metric : Literal["cosine", "dot"]
        Whether to compute the cosine similarity between the input and output embeddings
        or the dot product.
    rescale: Optional[float]
        Rescale the output cosine similarities by a constant factor.
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        num_groups: int = 0,
        metric: Literal["cosine", "dot"] = "cosine",
        rescale: Optional[float] = None,
        bias: bool = True,
    ):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = torch.nn.Parameter(torch.zeros(out_features, in_features))
        self.register_buffer(
            "groups", torch.zeros(num_groups, out_features, dtype=torch.bool)
        )
        self.rescale: float = (
            rescale if rescale is not None else 20.0 if metric == "cosine" else 1.0
        )
        self.metric = metric
        self.register_parameter(
            "bias",
            torch.nn.Parameter(torch.tensor(-0.65 if metric == "cosine" else 0.0))
            if bias
            else None,
        )
        self.reset_parameters()

        self._last_version = None
        self._normalized_weight = None

    def reset_parameters(self):
        torch.nn.init.xavier_uniform_(self.weight)

    def normalized_weight(self):
        if (
            (self.weight._version, id(self.weight)) == self._last_version
            and not self.training
            and self._normalized_weight is not None
        ):
            return self._normalized_weight
        normalized_weight = self.normalize_embedding(self.weight)
        if not self.training and normalized_weight is not self.weight:
            self._normalized_weight = normalized_weight
            self._last_version = (self.weight._version, id(self.weight))
        return normalized_weight

    def normalize_embedding(self, inputs):
        if self.metric == "cosine":
            inputs = F.normalize(inputs, dim=-1)
        return inputs

    def forward(self, inputs, group_indices=None, **kwargs):
        x = F.linear(
            self.normalize_embedding(inputs),
            self.normalized_weight(),
        )
        if self.bias is not None:
            x += self.bias
        if self.rescale != 1.0:
            x *= self.rescale
        if group_indices is not None and len(self.groups):
            x = x.masked_fill(~self.groups[group_indices], -10000)
        return x

    def extra_repr(self):
        return "in_features={}, out_features={}, rescale={}, groups={}".format(
            self.in_features,
            self.out_features,
            float(self.rescale or 1.0),
            self.groups.shape[0] if self.groups is not None else None,
        )



import pytest

from edsnlp.utils.examples import parse_example
from edsnlp.utils.extensions import rgetattr

EXAMPLES = [
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '3', 'metastase': ['metastase']}>cancer métastasé au stade 3</ent> voire au <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '4', 'metastase': ['metastase']}>stade 4</ent>.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '3', 'metastase': 'metastase'}>cancer métastasé au stade 3</ent> voire au <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '4', 'metastase': 'metastase'}>stade 4</ent>.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '3', 'metastase': 'metastase'}>cancer métastasé au stade 3</ent> voire au <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '4', 'metastase': 'metastase'}>stade 4</ent>.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '3', 'metastase': ['metastase']}>cancer métastasé au stade 3</ent> voire au stade 4.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '3', 'metastase': 'metastase'}>cancer métastasé au stade 3</ent> voire au stade 4.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '3', 'metastase': 'metastase'}>cancer métastasé au stade 3</ent> voire au stade 4.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '4', 'metastase': ['metastase']}>cancer métastasé au stade 3 voire au stade 4</ent>.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '4', 'metastase': 'metastase'}>cancer métastasé au stade 3 voire au stade 4</ent>.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '4', 'metastase': 'metastase'}>cancer métastasé au stade 3 voire au stade 4</ent>.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': ['3', '4'], 'metastase': ['metastase']}>cancer métastasé au stade 3 voire au stade 4</ent>.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': ['3', '4'], 'metastase': 'metastase'}>cancer métastasé au stade 3 voire au stade 4</ent>.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': ['3', '4'], 'metastase': 'metastase'}>cancer métastasé au stade 3 voire au stade 4</ent>.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '3', 'metastase': ['metastase']}>cancer métastasé au stade 3</ent> voire au stade 4.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '3', 'metastase': 'metastase'}>cancer métastasé au stade 3</ent> voire au stade 4.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '3', 'metastase': 'metastase'}>cancer métastasé au stade 3</ent> voire au stade 4.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '4', 'metastase': ['metastase']}>cancer métastasé au stade 3 voire au stade 4</ent>.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '4', 'metastase': 'metastase'}>cancer métastasé au stade 3 voire au stade 4</ent>.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '4', 'metastase': 'metastase'}>cancer métastasé au stade 3 voire au stade 4</ent>.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un cancer métastasé au <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '3', 'metastase': ['metastase']}>stade 3</ent> voire au <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '4', 'metastase': ['metastase']}>stade 4</ent>.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un cancer métastasé au <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '3', 'metastase': 'metastase'}>stade 3</ent> voire au <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '4', 'metastase': 'metastase'}>stade 4</ent>.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un cancer métastasé au <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '3', 'metastase': 'metastase'}>stade 3</ent> voire au <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '4', 'metastase': 'metastase'}>stade 4</ent>.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un cancer métastasé au <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '3', 'metastase': ['metastase']}>stade 3</ent> voire au stade 4.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un cancer métastasé au <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '3', 'metastase': 'metastase'}>stade 3</ent> voire au stade 4.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un cancer métastasé au <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '3', 'metastase': 'metastase'}>stade 3</ent> voire au stade 4.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un cancer métastasé au stade 3 voire au <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '4', 'metastase': ['metastase']}>stade 4</ent>.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un cancer métastasé au stade 3 voire au <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '4', 'metastase': 'metastase'}>stade 4</ent>.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un cancer métastasé au stade 3 voire au <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '4', 'metastase': 'metastase'}>stade 4</ent>.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': ['3', '4'], 'metastase': ['metastase']}>cancer </ent>métastasé au stade 3 voire au stade 4.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': ['3', '4'], 'metastase': 'metastase'}>cancer </ent>métastasé au stade 3 voire au stade 4.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': ['3', '4'], 'metastase': 'metastase'}>cancer </ent>métastasé au stade 3 voire au stade 4.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '3', 'metastase': ['metastase']}>cancer </ent>métastasé au stade 3 voire au stade 4.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '3', 'metastase': 'metastase'}>cancer </ent>métastasé au stade 3 voire au stade 4.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '3', 'metastase': 'metastase'}>cancer </ent>métastasé au stade 3 voire au stade 4.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '4', 'metastase': ['metastase']}>cancer </ent>métastasé au stade 3 voire au stade 4.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '4', 'metastase': 'metastase'}>cancer </ent>métastasé au stade 3 voire au stade 4.
    """,  # noqa: E501
    """
    Le patient présente une métastasis sur un <ent label_='Cancer' _.source='Solide' _.assigned={'stage': '4', 'metastase': 'metastase'}>cancer </ent>métastasé au stade 3 voire au stade 4.
    """,  # noqa: E501
]

ALL_PARAMS = [
    (True, True, None, None),
    (True, True, None, "keep_first"),
    (True, True, None, "keep_last"),
    (True, True, "keep_first", None),
    (True, True, "keep_first", "keep_first"),
    (True, True, "keep_first", "keep_last"),
    (True, True, "keep_last", None),
    (True, True, "keep_last", "keep_first"),
    (True, True, "keep_last", "keep_last"),
    (True, False, None, None),
    (True, False, None, "keep_first"),
    (True, False, None, "keep_last"),
    (True, False, "keep_first", None),
    (True, False, "keep_first", "keep_first"),
    (True, False, "keep_first", "keep_last"),
    (True, False, "keep_last", None),
    (True, False, "keep_last", "keep_first"),
    (True, False, "keep_last", "keep_last"),
    (False, True, None, None),
    (False, True, None, "keep_first"),
    (False, True, None, "keep_last"),
    (False, True, "keep_first", None),
    (False, True, "keep_first", "keep_first"),
    (False, True, "keep_first", "keep_last"),
    (False, True, "keep_last", None),
    (False, True, "keep_last", "keep_first"),
    (False, True, "keep_last", "keep_last"),
    (False, False, None, None),
    (False, False, None, "keep_first"),
    (False, False, None, "keep_last"),
    (False, False, "keep_first", None),
    (False, False, "keep_first", "keep_first"),
    (False, False, "keep_first", "keep_last"),
    (False, False, "keep_last", None),
    (False, False, "keep_last", "keep_first"),
    (False, False, "keep_last", "keep_last"),
]


@pytest.mark.parametrize("params,example", list(zip(ALL_PARAMS, EXAMPLES)))
def test_contextual(blank_nlp, params, example):

    include_assigned, replace_entity, reduce_mode_stage, reduce_mode_metastase = params

    blank_nlp.add_pipe(
        "normalizer",
        config=dict(lowercase=True, accents=True, quotes=True, pollution=False),
    )

    terms = [
        "cancer",
        "tumeur",
    ]
    regex = [
        r"adeno(carcinom|[\s-]?k)",
        r"neoplas",
        r"melanom",
    ]
    benine = "benign|benin"
    stage = "stade (I{1,3}V?|[1234])"
    metastase = "(metasta)"
    cancer = dict(
        source="Solide",
        regex=regex,
        terms=terms,
        regex_attr="NORM",
        exclude=dict(
            regex=benine,
            window=3,
        ),
        assign=[
            dict(
                name="stage",
                regex=stage,
                window=(-10, 10),
                replace_entity=replace_entity,
                reduce_mode=reduce_mode_stage,
            ),
            dict(
                name="metastase",
                regex=metastase,
                window=10,
                reduce_mode=reduce_mode_metastase,
            ),
        ],
    )
    lymphome = dict(
        source="Lymphome",
        regex=["lymphom", "lymphangio"],
        regex_attr="NORM",
    )
    patterns = [cancer, lymphome]

    blank_nlp.add_pipe(
        "eds.contextual-matcher",
        name="Cancer",
        config=dict(
            label="Cancer",
            patterns=patterns,
            include_assigned=include_assigned,
        ),
    )

    text, entities = parse_example(example=example)

    doc = blank_nlp(text)

    assert len(doc.ents) == len(entities)

    for entity, ent in zip(entities, doc.ents):

        for modifier in entity.modifiers:

            assert (
                rgetattr(ent, modifier.key) == modifier.value
            ), f"{modifier.key} labels don't match."



from libcpp.vector cimport vector
from spacy.matcher.phrasematcher cimport PhraseMatcher
from spacy.structs cimport SpanC
from spacy.tokens.doc cimport Doc
from spacy.tokens.span cimport Span
from spacy.typedefs cimport attr_t


cdef class EDSPhraseMatcher(PhraseMatcher):
    cdef attr_t space_hash
    cdef attr_t excluded_hash

    cdef void find_matches(self, Doc doc, int start_idx, int end_idx, vector[SpanC] *matches) nogil



import time
from itertools import chain
from pathlib import Path

import pandas as pd
import pytest
from spacy.tokens import Doc

import edsnlp.data
import edsnlp.processing
from edsnlp.data.converters import get_current_tokenizer

docs = [
    {
        "note_id": 1234,
        "note_text": "This is a test.",
        "entities": [
            {
                "note_nlp_id": 0,
                "start_char": 0,
                "end_char": 4,
                "lexical_variant": "This",
                "note_nlp_source_value": "test",
                "negation": True,
            },
            {
                "note_nlp_id": 1,
                "start_char": 5,
                "end_char": 7,
                "lexical_variant": "is",
                "note_nlp_source_value": "test",
            },
        ],
    },
    {
        "note_id": 0,
        "note_text": "This is an empty document.",
        "entities": None,
    },
]


@pytest.mark.parametrize(
    "reader_format,reader_converter,backend,writer_format,writer_converter,worker_io",
    [
        ("pandas", "omop", "simple", "pandas", "omop", False),
        ("pandas", "omop", "multiprocessing", "pandas", "omop", False),
        ("pandas", "omop", "spark", "pandas", "omop", False),
        ("parquet", "omop", "simple", "parquet", "omop", False),
        ("parquet", "omop", "multiprocessing", "parquet", "omop", False),
        ("parquet", "omop", "spark", "parquet", "omop", False),
        ("parquet", "omop", "multiprocessing", "parquet", "omop", True),
        ("parquet", "omop", "spark", "parquet", "omop", True),
    ],
)
def test_end_to_end(
    reader_format,
    reader_converter,
    backend,
    writer_format,
    writer_converter,
    worker_io,
    nlp_eds,
    tmp_path,
):
    nlp = nlp_eds
    rsrc = Path(__file__).parent.parent.resolve() / "resources"
    if reader_format == "pandas":
        pandas_dataframe = pd.DataFrame(docs)
        data = edsnlp.data.from_pandas(
            pandas_dataframe,
            converter=reader_converter,
        )
    elif reader_format == "parquet":
        data = edsnlp.data.read_parquet(
            rsrc / "docs.pq",
            converter=reader_converter,
            read_in_worker=worker_io,
        )
    else:
        raise ValueError(reader_format)

    data = data.map_pipeline(nlp)

    data = data.set_processing(
        backend=backend,
        show_progress=True,
        chunk_size=2,
        batch_by="words",
        batch_size=2,
        sort_chunks=True,
    )

    if writer_format == "pandas":
        data.to_pandas(converter=writer_converter)
    elif writer_format == "parquet":
        if backend == "spark":
            with pytest.raises(ValueError):
                data.write_parquet(
                    tmp_path,
                    converter=writer_converter,
                    write_in_worker=worker_io,
                )
            data.write_parquet(
                tmp_path,
                converter=writer_converter,
                accumulate=False,
                write_in_worker=worker_io,
            )
        else:
            data.write_parquet(
                tmp_path,
                converter=writer_converter,
                write_in_worker=worker_io,
            )
    else:
        raise ValueError(writer_format)


def test_multiprocessing_backend(frozen_ml_nlp):
    texts = ["Ceci est un exemple", "Ceci est un autre exemple"]
    docs = list(
        frozen_ml_nlp.pipe(
            texts * 20,
            batch_size=2,
        ).set_processing(
            backend="multiprocessing",
            num_cpu_workers=-1,
            sort_chunks=True,
            chunk_size=2,
            batch_by="words",
            show_progress=True,
        )
    )
    assert len(docs) == 40


def error_pipe(doc: Doc):
    if doc._.note_id == "text-3":
        raise ValueError("error")
    return doc


def test_multiprocessing_gpu_stub(frozen_ml_nlp):
    text1 = "Ceci est un exemple"
    text2 = "Ceci est un autre exemple"
    list(
        frozen_ml_nlp.pipe(
            chain.from_iterable(
                [
                    text1,
                    text2,
                ]
                for i in range(5)
            ),
        ).set_processing(
            batch_size=2,
            num_gpu_workers=1,
            num_cpu_workers=1,
            gpu_worker_devices=["cpu"],
        )
    )


def simple_converter(obj):
    tok = get_current_tokenizer()
    doc = tok(obj["content"])
    doc._.note_id = obj["id"]
    return doc


def test_multiprocessing_rb_error(ml_nlp):
    text1 = "Ceci est un exemple"
    text2 = "Ceci est un autre exemple"
    ml_nlp.add_pipe(error_pipe, name="error", after="sentences")
    with pytest.raises(ValueError):
        docs = edsnlp.data.from_iterable(
            chain.from_iterable(
                [
                    {"content": text1, "id": f"text-{i}"},
                    {"content": text2, "id": f"other-text-{i}"},
                ]
                for i in range(5)
            ),
            converter=simple_converter,
        ).map(lambda x: time.sleep(0.2) or x)
        docs = ml_nlp.pipe(
            docs,
            n_process=2,
            batch_size=2,
        )
        list(docs)


try:
    import torch

    from edsnlp.core.torch_component import TorchComponent

    class DeepLearningError(TorchComponent):
        def __init__(self, *args, **kwargs):
            super().__init__()

        def preprocess(self, doc):
            return {"num_words": len(doc), "doc_id": doc._.note_id}

        def collate(self, batch):
            return {
                "num_words": torch.tensor(batch["num_words"]),
                "doc_id": batch["doc_id"],
            }

        def forward(self, batch):
            if "text-1" in batch["doc_id"]:
                raise RuntimeError("Deep learning error")
            return {}

except (ImportError, AttributeError):
    pass


def test_multiprocessing_ml_error(ml_nlp):
    text1 = "Ceci est un exemple"
    text2 = "Ceci est un autre exemple"
    ml_nlp.add_pipe(
        DeepLearningError(pipeline=ml_nlp),
        name="error",
        after="sentences",
    )

    with pytest.raises(RuntimeError) as e:
        docs = edsnlp.data.from_iterable(
            chain.from_iterable(
                [
                    {"content": text1, "id": f"text-{i}"},
                    {"content": text2, "id": f"other-text-{i}"},
                ]
                for i in range(5)
            ),
            converter=simple_converter,
        )
        docs = ml_nlp.pipe(docs)
        docs = docs.set_processing(
            batch_size=2,
            num_gpu_workers=1,
            num_cpu_workers=1,
            gpu_worker_devices=["cpu"],
        )
        list(docs)
    assert "Deep learning error" in str(e.value)



"""`eds.diabetes` pipeline"""
from typing import Any, Dict, List, Optional, Union

from spacy.tokens import Doc, Span

from edsnlp.core import PipelineProtocol
from edsnlp.matchers.regex import RegexMatcher
from edsnlp.matchers.utils import get_text
from edsnlp.pipes.base import SpanSetterArg
from edsnlp.pipes.core.contextual_matcher.contextual_matcher import (
    get_window,
)

from ..base import DisorderMatcher
from .patterns import COMPLICATIONS, default_patterns


class DiabetesMatcher(DisorderMatcher):
    """
    The `eds.diabetes` pipeline component extracts mentions of diabetes.

    ??? info "Details of the used patterns"
        ```{ .python .no-check }
        # fmt: off
        --8<-- "edsnlp/pipes/ner/disorders/diabetes/patterns.py"
        # fmt: on
        ```

    Extensions
    ----------
    On each span `span` that match, the following attributes are available:

    - `span._.detailed_status`: set to either
        - `"WITH_COMPLICATION"` if the diabetes is  complicated (e.g., via organ
           damages)
        - `"WITHOUT_COMPLICATION"` otherwise
    - `span._.assigned`: dictionary with the following keys, if relevant:
        - `type`: type of diabetes (I or II)
        - `insulin`: if the diabetes is insulin-dependent
        - `corticoid`: if the diabetes is corticoid-induced

    Examples
    --------
    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.sentences())
    nlp.add_pipe(
        eds.normalizer(
            accents=True,
            lowercase=True,
            quotes=True,
            spaces=True,
            pollution=dict(
                information=True,
                bars=True,
                biology=True,
                doctors=True,
                web=True,
                coding=True,
                footer=True,
            ),
        ),
    )
    nlp.add_pipe(eds.diabetes())
    ```

    Below are a few examples:

    --8<-- "docs/assets/fragments/diabetes-examples.md"

    Parameters
    ----------
    nlp : Optional[PipelineProtocol]
        The pipeline
    name : Optional[str]
        The name of the component
    patterns: Union[Dict[str, Any], List[Dict[str, Any]]]
        The patterns to use for matching
    label : str
        The label to use for the `Span` object and the extension
    span_setter : SpanSetterArg
        The span setter to use

    # Authors and citation

    The `eds.diabetes` component was developed by AP-HP's Data Science team with a team
    of medical experts. A paper describing in details the development of those
    components is being drafted and will soon be available.
    """

    def __init__(
        self,
        nlp: Optional[PipelineProtocol],
        name: str = "diabetes",
        *,
        patterns: Union[Dict[str, Any], List[Dict[str, Any]]] = default_patterns,
        label: str = "diabetes",
        span_setter: SpanSetterArg = {"ents": True, "diabetes": True},
    ):
        super().__init__(
            nlp=nlp,
            name=name,
            patterns=patterns,
            detailed_status_mapping={
                0: "ABSENT",
                1: "WITHOUT_COMPLICATION",
                2: "WITH_COMPLICATION",
            },
            label=label,
            span_setter=span_setter,
        )

        self.complication_matcher = RegexMatcher(
            attr="NORM", ignore_excluded=True, alignment_mode="expand"
        )
        self.complication_matcher.build_patterns(
            regex=dict(far_complications=COMPLICATIONS)
        )

    def process(self, doc: Doc) -> List[Span]:
        for span in super().process(doc):
            if span._.source == "complicated":
                span._.status = 2

            elif any([k.startswith("complicated") for k in span._.assigned.keys()]):
                span._.status = 2

            elif (
                get_text(span, "NORM", ignore_excluded=True) == "db"
            ) and not span._.assigned:
                # Mostly FP
                continue

            elif self.has_far_complications(span):
                span._.status = 2

            yield span

    def has_far_complications(self, span: Span):
        """
        Handles the common case where complications are listed as bullet points,
        sometimes fairly far from the anchor.
        """
        window = (0, 50)
        context = get_window(span, window, limit_to_sentence=False)
        if next(iter(self.complication_matcher(context)), None) is not None:
            return True
        return False



from edsnlp.core import registry

from .reported_speech import ReportedSpeechQualifier

DEFAULT_CONFIG = dict(
    pseudo=None,
    preceding=None,
    following=None,
    quotation=None,
    verbs=None,
    attr="NORM",
    span_getter=None,
    on_ents_only=True,
    within_ents=False,
    explain=False,
)

create_component = registry.factory.register(
    "eds.reported_speech",
    assigns=["span._.reported_speech"],
    deprecated=[
        "reported_speech",
        "rspeech",
    ],
)(ReportedSpeechQualifier)



"""`eds.adicap` pipeline"""
import re
from typing import List, Optional, Union

from spacy.tokens import Doc, Span

from edsnlp.core import PipelineProtocol
from edsnlp.pipes.base import SpanSetterArg
from edsnlp.pipes.core.contextual_matcher import ContextualMatcher
from edsnlp.utils.resources import get_adicap_dict

from . import patterns
from .models import AdicapCode


# noinspection SpellCheckingInspection
class AdicapMatcher(ContextualMatcher):
    """
    The `eds.adicap` pipeline component matches the ADICAP codes. It was developped to
    run on anapathology reports.

    !!! warning "Document type"

        It was developped to work on anapathology reports.
        We recommend also to use the `eds` language (`edsnlp.blank("eds")`)

    The compulsory characters of the ADICAP code are identified and decoded.
    These characters represent the following attributes:

    | Field [en]        | Field [fr]                    | Attribute       |
    |-------------------|-------------------------------|-----------------|
    | Sampling mode     | Mode de prelevement           | sampling_mode   |
    | Technic           | Type de technique             | technic         |
    | Organ and regions | Appareils, organes et régions | organ           |
    | Pathology         | Pathologie générale           | pathology       |
    | Pathology type    | Type de la pathologie         | pathology_type  |
    | Behaviour type    | Type de comportement          | behaviour_type  |


    The pathology field takes 4 different values corresponding to the 4 possible
    interpretations of the ADICAP code, which are : "PATHOLOGIE GÉNÉRALE NON TUMORALE",
    "PATHOLOGIE TUMORALE", "PATHOLOGIE PARTICULIERE DES ORGANES" and "CYTOPATHOLOGIE".

    Depending on the pathology value the behaviour type meaning changes, when the
    pathology is tumoral then it describes the malignancy of the tumor.

    For further details about the ADICAP code follow this [link](https://smt.esante.\
gouv.fr/wp-json/ans/terminologies/document?terminologyId=terminologie-adicap&file\
Name=cgts_sem_adicap_fiche-detaillee.pdf).

    Examples
    --------
    ```{ .python .no-check }
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.sentences())
    nlp.add_pipe(eds.adicap())

    text = \"\"\"
    COMPTE RENDU D’EXAMEN

    Antériorité(s) :  NEANT


    Renseignements cliniques :
    Contexte d'exploration d'un carcinome canalaire infiltrant du quadrant supéro-
    externe du sein droit. La lésion biopsiée ce jour est située à 5,5 cm de la lésion
    du quadrant supéro-externe, à l'union des quadrants inférieurs.


    Macrobiopsie 10G sur une zone de prise de contraste focale à l'union des quadrants
    inférieurs du sein droit, mesurant 4 mm, classée ACR4

    14 fragments ont été communiqués fixés en formol (lame n° 1a et lame n° 1b) . Il
    n'y a pas eu d'échantillon congelé. Ces fragments ont été inclus en paraffine en
    totalité et coupés sur plusieurs niveaux.
    Histologiquement, il s'agit d'un parenchyme mammaire fibroadipeux parfois
    légèrement dystrophique avec quelques petits kystes. Il n'y a pas d'hyperplasie
    épithéliale, pas d'atypie, pas de prolifération tumorale. On note quelques
    suffusions hémorragiques focales.

    Conclusion :
    Légers remaniements dystrophiques à l'union des quadrants inférieurs du sein droit.
    Absence d'atypies ou de prolifération tumorale.

    Codification :   BHGS0040
    \"\"\"

    doc = nlp(text)

    doc.ents
    # Out: (BHGS0040,)

    ent = doc.ents[0]

    ent.label_
    # Out: adicap

    ent._.adicap.dict()
    # Out: {'code': 'BHGS0040',
    # 'sampling_mode': 'BIOPSIE CHIRURGICALE',
    # 'technic': 'HISTOLOGIE ET CYTOLOGIE PAR INCLUSION',
    # 'organ': "SEIN (ÉGALEMENT UTILISÉ CHEZ L'HOMME)",
    # 'pathology': 'PATHOLOGIE GÉNÉRALE NON TUMORALE',
    # 'pathology_type': 'ETAT SUBNORMAL - LESION MINEURE',
    # 'behaviour_type': 'CARACTERES GENERAUX'}
    ```

    Parameters
    ----------
    nlp : Optional[PipelineProtocol]
        The pipeline object
    name : str
        The name of the pipe
    pattern : Optional[Union[List[str], str]]
        The regex pattern to use for matching ADICAP codes
    prefix : Optional[Union[List[str], str]]
        The regex pattern to use for matching the prefix before ADICAP codes
    window : int
        Number of tokens to look for prefix. It will never go further the start of
        the sentence
    attr : str
        Attribute to match on, eg `TEXT`, `NORM`, etc.
    label : str
        Label name to use for the `Span` object and the extension
    span_setter : SpanSetterArg
        How to set matches on the doc

    Authors and citation
    --------------------
    The `eds.adicap` pipeline was developed by AP-HP's Data Science team.
    The codes were downloaded from the website of 'Agence du numérique en santé'
    ("Thésaurus de la codification ADICAP - Index raisonné des
    lésions", [@terminologie-adicap])
    """

    def __init__(
        self,
        nlp: Optional[PipelineProtocol],
        name: str = "adicap",
        *,
        pattern: Union[List[str], str] = patterns.base_code,
        prefix: Union[List[str], str] = patterns.adicap_prefix,
        window: int = 500,
        attr: str = "TEXT",
        label: str = "adicap",
        span_setter: SpanSetterArg = {"ents": True, "adicap": True},
    ):
        adicap_pattern = dict(
            source="adicap",
            regex=prefix,
            regex_attr=attr,
            assign=[
                dict(
                    name="code",
                    regex=pattern,
                    window=window,
                    replace_entity=True,
                    reduce_mode=None,
                ),
            ],
        )

        super().__init__(
            nlp=nlp,
            name=name,
            label=label,
            attr=attr,
            patterns=adicap_pattern,
            ignore_excluded=False,
            regex_flags=0,
            alignment_mode="expand",
            include_assigned=False,
            assign_as_span=False,
            span_setter=span_setter,
        )

        self.decode_dict = get_adicap_dict()

        self.set_extensions()

    def set_extensions(self) -> None:
        super().set_extensions()
        if not Span.has_extension(self.label):
            Span.set_extension(self.label, default=None)

    def decode(self, code):
        code = re.sub("[^A-Za-z0-9 ]+", "", code)
        exploded = list(code)
        adicap = AdicapCode(
            code=code,
            sampling_mode=self.decode_dict["D1"]["codes"].get(exploded[0]),
            technic=self.decode_dict["D2"]["codes"].get(exploded[1]),
            organ=self.decode_dict["D3"]["codes"].get("".join(exploded[2:4])),
        )

        for d in ["D4", "D5", "D6", "D7"]:
            adicap_short = self.decode_dict[d]["codes"].get("".join(exploded[4:8]))
            adicap_long = self.decode_dict[d]["codes"].get("".join(exploded[2:8]))

            if (adicap_short is not None) | (adicap_long is not None):
                adicap.pathology = self.decode_dict[d]["label"]
                adicap.behaviour_type = self.decode_dict[d]["codes"].get(exploded[5])

                if adicap_short is not None:
                    adicap.pathology_type = adicap_short

                else:
                    adicap.pathology_type = adicap_long

        return adicap

    def process(self, doc: Doc) -> List[Span]:
        """
        Tags ADICAP mentions.

        Parameters
        ----------
        doc : Doc
            spaCy Doc object

        Returns
        -------
        doc : Doc
            spaCy Doc object, annotated for ADICAP
        """
        for span in super().process(doc):
            span._.set(self.label, self.decode(span._.assigned["code"]))
            yield span



results_peptic_ulcer_disease = dict(
    has_match=[
        True,
        True,
        False,
        True,
    ],
    detailled_status="PRESENT",
    assign=None,
    texts=[
        "Beaucoup d'ulcères gastriques",
        "Présence d'UGD",
        "La patient à des ulcères",
        "Au niveau gastrique: " + 5 * "blabla " + "quelques ulcères",
    ],
)



from edsnlp.core import registry

from .spaces import SpacesTagger

DEFAULT_CONFIG = dict(newline=True)

create_component = registry.factory.register(
    "eds.spaces",
    assigns=["token.tag"],
    deprecated=["spaces"],
)(SpacesTagger)



from edsnlp.utils.examples import parse_example

examples = [
    "TNM: <ent norm=aTxN1M0>aTxN1M0</ent>",
    "TNM: <ent norm=pTxN1M0>p Tx N1M 0</ent>",
    "TNM: <ent norm='pTxN1M0 (UICC 2020)'>p Tx N1M 0 (UICC 20)</ent>",
    "TNM: <ent norm='aTxN1M0 (UICC 1968)'>aTxN1M0 (UICC 68)</ent>",
    "TNM: <ent norm=aTxN1R2>aTxN1 R2</ent>",
    "TNM: <ent norm='pT2cN0R0 (TNM 2010)'>pT2c N0 R0 (TNM 2010)</ent>",
    "TNM: <ent norm=aTxN1M0>aTx / N1 / M0</ent>",
    "TNM: <ent norm=pT2N1mi>pT2 N1mi</ent>",
    "TNM: <ent norm=pT1mN1M0>pT1(m)N1 M0</ent>",
    "TNM: <ent norm=pT1bN0sn>pT1bN0(sn)</ent>",
    "TNM: <ent norm=pT1N1M0>pT1 pN1 M0</ent>\n \n ",
    "TNM: <ent norm=aTxN1M0>aTxN1M0</ent> ",
    "TNM: <ent norm=cT3N0M0>cT3N0M0</ent> \n \n",
    "TNM: PT",
    "TNM: p    T \n",
    "TNM: a T \n",
    "TNM: <ent norm=p>pT</ent> \n  \n0",
]


def test_scores(blank_nlp):
    blank_nlp.add_pipe("eds.tnm")

    for example in examples:
        text, entities = parse_example(example=example)

        doc = blank_nlp(text)

        assert len(entities) == len(doc.ents)

        for entity, ent in zip(entities, doc.ents):
            norm = entity.modifiers[0].value
            assert ent.text == text[entity.start_char : entity.end_char]
            assert norm == ent._.value.norm()



import pytest
import spacy
from AIDS import results_aids
from alcohol import results_alcohol
from cerebrovascular_accident import results_cerebrovascular_accident
from CKD import results_ckd
from congestive_heart_failure import results_congestive_heart_failure
from connective_tissue_disease import results_connective_tissue_disease
from COPD import results_copd
from dementia import results_dementia
from diabetes import results_diabetes
from hemiplegia import results_hemiplegia
from leukemia import results_leukemia
from liver_disease import results_liver_disease
from lymphoma import results_lymphoma
from myocardial_infarction import results_myocardial_infarction
from peptic_ulcer_disease import results_peptic_ulcer_disease
from peripheral_vascular_disease import results_peripheral_vascular_disease
from solid_tumor import results_solid_tumor
from tobacco import results_tobacco

results = dict(
    aids=results_aids,
    ckd=results_ckd,
    copd=results_copd,
    alcohol=results_alcohol,
    cerebrovascular_accident=results_cerebrovascular_accident,
    congestive_heart_failure=results_congestive_heart_failure,
    connective_tissue_disease=results_connective_tissue_disease,
    dementia=results_dementia,
    diabetes=results_diabetes,
    hemiplegia=results_hemiplegia,
    leukemia=results_leukemia,
    liver_disease=results_liver_disease,
    lymphoma=results_lymphoma,
    myocardial_infarction=results_myocardial_infarction,
    peptic_ulcer_disease=results_peptic_ulcer_disease,
    peripheral_vascular_disease=results_peripheral_vascular_disease,
    solid_tumor=results_solid_tumor,
    tobacco=results_tobacco,
)


@pytest.fixture
def normalized_nlp(lang):
    model = spacy.blank(lang)
    model.add_pipe("eds.sentences")
    model.add_pipe("eds.normalizer")
    return model


class DisorderTester:
    def __init__(
        self,
        disorder,
        nlp,
        texts,
        has_match,
        detailled_status,
        assign=None,
    ):
        self.disorder = disorder
        self.nlp = nlp

        self.texts = texts

        self.has_match = has_match
        self.detailled_status = (
            detailled_status
            if isinstance(detailled_status, list)
            else len(texts) * [detailled_status]
        )
        self.assign = assign if assign is not None else len(texts) * [None]

        self.nlp.add_pipe(f"eds.{disorder}")

    def check(self):
        for text, has_match, detailled_status, assign in zip(
            self.texts, self.has_match, self.detailled_status, self.assign
        ):
            doc = self.nlp(text)
            ents = doc.spans[self.disorder]

            assert len(ents) == int(has_match)

            for ent in ents:
                assert ent.label_ == self.disorder

            if not ents:
                continue

            ent = ents[0]

            assert ent._.detailled_status == detailled_status

            if assign is not None:
                for key, value in assign.items():
                    assert repr(value) == repr(assign[key])


@pytest.mark.parametrize(
    "disorder",
    list(results.keys()),
)
def test_disorder(normalized_nlp, disorder):

    result = results[disorder]

    expect = DisorderTester(
        disorder,
        normalized_nlp,
        **result,
    )

    expect.check()


def test_behavior_negation(blank_nlp):
    blank_nlp.add_pipe("eds.normalizer")
    blank_nlp.add_pipe("eds.tobacco")
    blank_nlp.add_pipe("eds.alcohol")
    blank_nlp.add_pipe("eds.negation")

    negated_texts = [
        "Le patient ne fume aucun truc.",
        "Le patient fume 0 PA.",
        "Il ne boit pas d'alcool." "Boit très rarement de l'alcool.",
    ]

    positive_texts = ["Le patient ne fume que le soir.", "Il ne boit plus d'alcool."]

    for text in negated_texts:
        doc = blank_nlp(text)
        assert doc.ents[0]._.negation is True, text

    for text in positive_texts:
        doc = blank_nlp(text)
        assert doc.ents[0]._.negation is False, text



main_pattern = dict(
    source="main",
    regex=[
        r"alveolites.{1,5}fibrosante",
        r"asthm",
        r"broncho.?pneumopathies.{1,5}chroniques.{1,5}obstru",
        r"bronchites.{1,5}chroniques.{1,5}obstru",
        r"fibro.{1,20}(poumon|pulmo|pleur)",
        r"fibrose.{1,5}interstitielle.{1,5}diffuse.{1,5}idiopathique",
        r"fibrose.{1,5}intersti",
        r"obstruction.{1,5}chronique.{1,10}voie.{1,5}aerienne",
        r"pneumoconiose",
        r"pneumo(nie|pathie).{0,15}(intersti|radiq|infiltr|fibro|organis)",
        r"poumon.{1,5}noir",
        r"sclerose.{1,5}pulmo",
        r"fibro.?elastose.{1,5}pleuro.?paren",
        r"apnee.{1,25}sommeil",
        r"emphyseme",
        r"insuffisan.{1,5}respiratoire.{1,5}chron",
        r"mucoviscidose",
        r"bronchiolite.oblilerante.{1,10}pneumo.{1,20}organis",
    ],
    regex_attr="NORM",
)

htap = dict(
    source="htap",
    regex=[
        r"\bhtap\b",
        r"hypertension.{0,10}pulmo",
        r"hypertension.{1,5}arter.{1,15}(poumon|pulmo)",
    ],
    regex_attr="NORM",
    exclude=[
        dict(
            regex="minime",
            window=(0, 3),
        ),
    ],
)

oxygen = dict(
    source="oxygen",
    regex=[
        r"oxygeno.?dependance",
        r"oxygeno.?requeran",
        r"oxygenation",
        r"oxygeno.?ther",
        r"oxygene",
    ],
    regex_attr="NORM",
    assign=[
        dict(
            name="long",
            regex=r"(long.{1,10}(?:cour|dure)|chroni|domicil)",
            window=6,
        ),
        dict(
            name="long_bis",
            regex=r"(persist|major|minor)",
            window=-6,
        ),
        dict(
            name="need",
            regex=r"(besoin)",
            window=(-6, 6),
        ),
    ],
)

acronym = dict(
    source="acronym",
    regex=[
        r"\bBPCO\b",
        r"\bFPI\b",
        r"\bOLD\b",
        r"\bFEPP\b",
        r"\bPINS\b",
        r"\bPID\b",
        r"\bSAOS\b",
        r"\bSAS\b",
        r"\bSAHOS\b",
        r"\bBOOP\b",
    ],
    regex_attr="TEXT",
)

fid = dict(
    source="fid",
    regex=[
        r"\bfid\b",
    ],
    regex_attr="NORM",
    exclude=[
        dict(
            regex=[
                r"\bfig\b",
                r"palpation",
            ],
            window=(-7, 7),
        ),
    ],
)

default_patterns = [
    main_pattern,
    htap,
    oxygen,
    acronym,
    fid,
]



from typing import Optional, Sequence

import torch
import torch.nn.functional as F
from typing_extensions import Literal

from edsnlp.utils.torch import ActivationFunction, get_activation_function


class Residual(torch.nn.Module):
    def __init__(self, normalize: Literal["pre", "post", "none"] = "pre"):
        super().__init__()
        self.normalize = normalize

    def forward(self, before, after):
        return (
            before + F.layer_norm(after, after.shape[-1:])
            if self.normalize == "pre"
            else F.layer_norm(before + after, after.shape[-1:])
            if self.normalize == "post"
            else before + after
        )


class TextCnn(torch.nn.Module):
    def __init__(
        self,
        input_size: int,
        output_size: Optional[int] = None,
        out_channels: Optional[int] = None,
        kernel_sizes: Sequence[int] = (3, 4, 5),
        activation: ActivationFunction = "relu",
        residual: bool = True,
        normalize: Literal["pre", "post", "none"] = "pre",
    ):
        """
        Parameters
        ----------
        input_size: int
            Size of the input embeddings
        output_size: Optional[int]
            Size of the output embeddings
            Defaults to the `input_size`
        out_channels: int
            Number of channels
        kernel_sizes: Sequence[int]
            Window size of each kernel
        activation: str
            Activation function to use
        residual: bool
            Whether to use residual connections
        normalize: Literal["pre", "post", "none"]
            Whether to normalize before or after the residual connection
        """
        super().__init__()

        if out_channels is None:
            out_channels = input_size
        output_size = input_size if output_size is None else output_size

        self.convolutions = torch.nn.ModuleList(
            torch.nn.Conv1d(
                in_channels=input_size,
                out_channels=out_channels,
                kernel_size=kernel_size,
                padding=0,
            )
            for kernel_size in kernel_sizes
        )
        self.linear = torch.nn.Linear(
            in_features=out_channels * len(kernel_sizes),
            out_features=output_size,
        )
        self.activation = get_activation_function(activation)
        self.residual = Residual(normalize=normalize) if residual else None

    def forward(
        self, embeddings: torch.FloatTensor, mask: torch.BoolTensor
    ) -> torch.FloatTensor:
        # shape: samples words dim
        max_k = max(conv.kernel_size[0] for conv in self.convolutions)
        left_pad = (max_k) // 2
        right_pad = (max_k - 1) // 2
        n_samples, n_words, dim = embeddings.shape
        n_words_with_pad = n_words + left_pad + right_pad

        # shape: samples (left_pad... words ...right_pad) dim
        padded_x = F.pad(embeddings, pad=(0, 0, max_k // 2, (max_k - 1) // 2))
        padded_mask = F.pad(mask, pad=(max_k // 2 + (max_k - 1) // 2, 0), value=True)

        # shape: (un-padded sample words) dim
        flat_x = padded_x[padded_mask]

        # Conv-1d expects sample * dim * words
        flat_x = flat_x.permute(1, 0).unsqueeze(0)

        # Apply the convolutions over the flattened input
        conv_results = []
        for conv_idx, conv in enumerate(self.convolutions):
            k = conv.kernel_size[0]
            conv_x = conv(flat_x)
            offset_left = left_pad - (k // 2)
            offset_right = conv_x.size(2) - (right_pad - ((k - 1) // 2))
            conv_results.append(conv_x[0, :, offset_left:offset_right])
        flat_x = torch.cat(conv_results, dim=0)
        flat_x = flat_x.transpose(1, 0)  # n_words * dim

        # Apply the non-linearities
        flat_x = torch.relu(flat_x)
        flat_x = self.linear(flat_x)

        # Reshape the output to the original shape
        new_dim = flat_x.size(-1)
        x = torch.empty(n_samples * n_words_with_pad, new_dim, device=flat_x.device)
        flat_mask = padded_mask.clone()
        flat_mask[-1, padded_mask[-1].sum() - right_pad :] = False
        flat_mask[0, :left_pad] = False
        flat_mask = flat_mask.view(-1)
        x[flat_mask] = flat_x
        x = x.view(n_samples, n_words_with_pad, new_dim)
        x = x[:, left_pad:-right_pad]

        # Apply the residual connection
        if self.residual is not None:
            x = self.residual(embeddings, x)

        return x.masked_fill_((~mask).unsqueeze(-1), 0)



def test_drugs(blank_nlp):
    blank_nlp.add_pipe("eds.normalizer")
    blank_nlp.add_pipe("eds.drugs")

    text = "Traitement habituel: Kardégic, cardensiel (bisoprolol), glucophage, lasilix"
    doc = blank_nlp(text)
    drugs_expected = [
        ("Kardégic", "B01AC06"),
        ("cardensiel", "C07AB07"),
        ("bisoprolol", "C07AB07"),
        ("glucophage", "A10BA02"),
        ("lasilix", "C03CA01"),
    ]
    drugs_detected = [(x.text, x.kb_id_) for x in doc.ents]
    assert drugs_detected == drugs_expected



import json
import re
from typing import Any, Dict, List, Tuple, Union

import regex
from pydantic import BaseModel, validator


class Match(BaseModel):
    start_char: int
    end_char: int
    text: str
    modifiers: str


class Modifier(BaseModel):
    key: str
    value: Union[int, float, bool, str, Dict[str, Any]]

    @validator("value")
    def optional_dict_parsing(cls, v):
        if v in ["True", "False"]:
            return v == "True"
        if isinstance(v, str):
            try:
                return json.loads(v.replace("'", '"'))
            except json.JSONDecodeError:
                return v
        return v


class Entity(BaseModel):
    start_char: int
    end_char: int
    modifiers: List[Modifier]


entity_pattern = re.compile(r"(<ent[^<>]*>[^<>]+</ent>)", flags=re.DOTALL)
text_pattern = re.compile(r"<ent.*>(.+)</ent>", flags=re.DOTALL)
modifiers_pattern = re.compile(r"<ent\s?(.*)>.+</ent>", flags=re.DOTALL)
single_modifiers_pattern = regex.compile(
    r"(?P<key>[^\s]+?)=((?P<value>{.*?})|(?P<value>[^\s']+)|'(?P<value>.+?)')",
    flags=regex.DOTALL,
)


def find_matches(example: str) -> List[re.Match]:
    """
    Finds entities within the example.

    Parameters
    ----------
    example : str
        Example to process.

    Returns
    -------
    List[re.Match]
        List of matches for entities.
    """
    return list(entity_pattern.finditer(example))


def parse_match(match: re.Match) -> Match:
    """
    Parse a regex match representing an entity.

    Parameters
    ----------
    match : re.Match
        Match for an entity.

    Returns
    -------
    Match
        Usable representation for the entity match.
    """

    lexical_variant = match.group()
    start_char = match.start()
    end_char = match.end()

    text = text_pattern.findall(lexical_variant)[0]
    modifiers = modifiers_pattern.findall(lexical_variant)[0]

    m = Match(start_char=start_char, end_char=end_char, text=text, modifiers=modifiers)

    return m


def parse_example(example: str) -> Tuple[str, List[Entity]]:
    """
    Parses an example : finds examples and removes the tags.

    Parameters
    ----------
    example : str
        Example to process.

    Returns
    -------
    Tuple[str, List[Entity]]
        Cleaned text and extracted entities.
    """

    matches = [parse_match(match) for match in find_matches(example=example)]
    text = ""
    entities = []

    cursor = 0

    for match in matches:
        text += example[cursor : match.start_char]
        start_char = len(text)
        text += match.text
        end_char = len(text)

        cursor = match.end_char

        entity = Entity(
            start_char=start_char,
            end_char=end_char,
            modifiers=[
                Modifier.parse_obj(m.groupdict())
                for m in single_modifiers_pattern.finditer(match.modifiers)
            ],
        )

        entities.append(entity)

    text += example[cursor:]

    return text, entities



from typing import Any, Callable, Dict, Iterable, Union, Sequence

from spacy.tokens import Doc
from spacy.training import Example

import numpy as np

Scorer = Union[
    Callable[[Iterable[Doc], Iterable[Doc]], Dict[str, Dict[str, Any]]],
    Callable[[Iterable[Example]], Dict[str, Dict[str, Any]]],
]


def average_precision(pred: Dict[Any, float], gold: Iterable[Any]):
    # Average precision computation (pred is {prediction -> probability})
    pred = sorted(pred, key=lambda k: pred[k], reverse=True)
    correct = [p in gold for p in pred]
    cum_correct = np.cumsum(correct)
    num_gold = len(gold)
    precisions = cum_correct / np.arange(1, len(correct) + 1)
    recalls = cum_correct / num_gold if num_gold > 0 else np.zeros(len(correct))
    ap = 0.0
    for i in range(1, len(precisions)):
        if recalls[i] > recalls[i - 1]:
            ap += (recalls[i] - recalls[i - 1]) * precisions[i]
    return ap


def prf(pred: Sequence, gold: Sequence):
    tp = len(set(pred) & set(gold))
    num_pred = len(pred)
    num_gold = len(gold)
    return {
        "f": 2 * tp / max(1, num_pred + num_gold),
        "p": 1 if tp == num_pred else (tp / num_pred),
        "r": 1 if tp == num_gold else (tp / num_gold),
        "tp": tp,
        "support": num_gold,  # num gold
        "positives": num_pred,  # num predicted
    }


def make_examples(*args):
    if len(args) == 2:
        return (
            [Example(reference=g, predicted=p) for g, p in zip(*args)]
            if len(args) == 2
            else args[0]
        )
    else:
        raise ValueError("Expected either a list of examples or two lists of spans")



from edsnlp.core import registry

from .dates import DatesMatcher

DEFAULT_CONFIG = dict(
    absolute=None,
    relative=None,
    duration=None,
    false_positive=None,
    on_ents_only=False,
    span_getter=None,
    merge_mode="intersect",
    detect_periods=False,
    detect_time=True,
    period_proximity_threshold=3,
    as_ents=False,
    attr="LOWER",
    date_label="date",
    duration_label="duration",
    period_label="period",
    span_setter={
        "dates": ["date"],
        "durations": ["duration"],
        "periods": ["period"],
    },
)

create_component = registry.factory.register(
    "eds.dates",
    assigns=["doc.spans", "doc.ents"],
    deprecated=["dates"],
)(DatesMatcher)



import re

import pytest
from helpers import make_nlp
from pytest import mark

from edsnlp.matchers.regex import RegexMatcher, create_span
from edsnlp.matchers.utils import get_text


def test_regex(doc):
    matcher = RegexMatcher()

    matcher.add("test", [r"test"])
    matcher.remove("test")

    matcher.add("patient", [r"patient"])

    matches = matcher(doc, as_spans=False)

    for _, start, end in matcher(doc, as_spans=False):
        assert len(doc[start:end])

    matches = matcher(doc[:10])

    assert list(matches)


@mark.parametrize(
    "pattern, txt, span_from_group, result",
    [
        (
            r"match1 (?:group1|(group2))",  # pattern
            "It is a match1 group1",  # txt
            True,  # span_from_group
            "match1 group1",  # result
        ),
        (
            r"match1 (?:group1|(group2))",
            "It is a match1 group1",
            False,
            "match1 group1",
        ),
        (
            r"match1 (?:group1|(group2))",
            "It is a match1 group2",
            True,
            "group2",
        ),
        (
            r"match1 (?:group1|(group2))",
            "It is a match1 group2",
            False,
            "match1 group2",
        ),
    ],
)
def test_regex_with_groups(blank_nlp, pattern, txt, span_from_group, result):
    doc = blank_nlp(txt)
    matcher = RegexMatcher(span_from_group=span_from_group)
    matcher.add("test", [pattern])
    match = list(matcher(doc, as_spans=True))[0].text
    assert match == result


def test_regex_with_norm(blank_nlp):
    blank_nlp.add_pipe("pollution")

    text = "pneumopathie à NBNbWbWbNbWbNBNbNbWbWbNBNbWbNbNbWbNBNbWbNbNB coronavirus"

    doc = blank_nlp(text)

    matcher = RegexMatcher(ignore_excluded=True)
    matcher.add("test", ["pneumopathie à coronavirus"])

    match = list(matcher(doc, as_spans=True))[0]
    assert match.text == text
    assert match._.normalized_variant == "pneumopathie à coronavirus"


def test_regex_with_norm_on_span(blank_nlp):
    blank_nlp.add_pipe("pollution")

    text = (
        "le patient a une pneumopathie à NBNbWbWbNbWbNBNbNbWbWbNBNbWbNbNbWbNBNbWbNbNB"
        " coronavirus"
    )

    for offset in (0, 2):
        doc = blank_nlp(text)[offset:]

        matcher = RegexMatcher(ignore_excluded=True)
        matcher.add("test", ["pneumopathie à coronavirus"])

        match = list(matcher(doc, as_spans=True))[0]
        assert (
            match.text
            == "pneumopathie à NBNbWbWbNbWbNBNbNbWbWbNBNbWbNbNbWbNBNbWbNbNB coronavirus"
        )
        assert match._.normalized_variant == "pneumopathie à coronavirus"


def test_offset(blank_nlp):
    text = "Ceci est un test de matching"

    doc = blank_nlp(text)
    pattern = "matching"

    matcher = RegexMatcher(attr="TEXT")

    matcher.add("test", [pattern])

    for _, start, end in matcher(doc):
        assert doc[start:end].text == pattern

    for span in matcher(doc, as_spans=True):
        span.text == pattern

    for _, start, end in matcher(doc[2:]):
        assert doc[2:][start:end].text == pattern

    for span in matcher(doc[2:], as_spans=True):
        span.text == pattern


def test_remove():
    matcher = RegexMatcher(attr="TEXT")

    matcher.add("test", ["pattern"])
    matcher.add("test", ["pattern2"], attr="LOWER")

    assert len(matcher) == 1

    with pytest.raises(ValueError):
        matcher.remove("wrong_key")

    matcher.remove("test")

    assert len(matcher) == 0


def test_norm_alignment(blank_nlp):
    text = "test " + "bla… " * 4 + "test " + "bla" * 10

    blank_nlp.add_pipe(
        "matcher", config=dict(regex=dict(test=r"\btest\b"), attr="NORM")
    )

    doc = blank_nlp(text)

    for ent in doc.ents:
        assert ent.text == "test"


@mark.parametrize(
    "leading_text",
    [
        "",
        "\n",
        "Test de non-pollution",
    ],
)
@mark.parametrize("leading_pollution", [True, False])
@mark.parametrize("pollution_within", [True, False])
@mark.parametrize("trailing_pollution", [True, False])
@mark.parametrize(
    "pollution",
    ["==================", "======= ======= =======", "Nnnnnnnnnnnnn nnnnnn nnnnnnnn"],
)
def text_get_text(
    blank_nlp,
    leading_text: str,
    leading_pollution: bool,
    pollution_within: bool,
    trailing_pollution: bool,
    pollution: str,
):
    if pollution_within:
        example = f"transplantation {pollution} cardiaque en 2000."
    else:
        example = "transplantation cardiaque en 2000."

    chunks = []

    if leading_text:
        chunks.append(leading_text)
    if leading_pollution:
        chunks.append(pollution)

    chunks.append(example)

    if trailing_pollution:
        chunks.append(pollution)

    text = " ".join(chunks)

    blank_nlp.add_pipe("eds.normalizer", config=dict(pollution=True))
    blank_nlp.add_pipe(
        "eds.matcher",
        config=dict(
            regex=dict(test="transplantation cardiaque"),
            attr="NORM",
            ignore_excluded=True,
        ),
    )
    doc = blank_nlp(text)

    clean = get_text(doc, attr="NORM", ignore_excluded=True)
    if leading_text:
        assert clean == f"{leading_text.lower()} transplantation cardiaque en 2000."
    else:
        assert clean == "transplantation cardiaque en 2000."

    assert doc.ents
    assert doc.ents[0][0].text == "transplantation"

    clean = get_text(doc.ents[0], attr="NORM", ignore_excluded=True)
    assert clean == "transplantation cardiaque"


def test_groupdict_as_spans(doc):
    matcher = RegexMatcher()

    matcher.add("test", [r"patient(?i:(?=.*(?P<cause>douleurs))?)"])

    [(span0, gd0), (span1, gd1)] = list(matcher.match_with_groupdict_as_spans(doc))
    assert span0.text == "patient"
    assert span1.text == "patient"
    assert len(gd0) == 1 and gd0["cause"].text == "douleurs"
    assert len(gd1) == 0


def test_regex_with_space(blank_nlp):
    blank_nlp.add_pipe("eds.spaces")

    text = "pneumopathie à      coronavirus"

    doc = blank_nlp(text)

    matcher = RegexMatcher(ignore_space_tokens=False)
    matcher.add("test", ["pneumopathie à coronavirus"])

    assert len(list(matcher(doc, as_spans=True))) == 0

    matcher = RegexMatcher(ignore_space_tokens=True)
    matcher.add("test", ["pneumopathie à coronavirus"])

    match = list(matcher(doc, as_spans=True))[0]
    assert match.text == text
    assert match._.normalized_variant == "pneumopathie à coronavirus"


@pytest.fixture(scope="session")
def doc2(lang):
    blank_nlp = make_nlp(lang)
    blank_nlp.add_pipe("eds.pollution")
    blank_nlp.add_pipe("eds.spaces")

    text = (
        "-----------------------------------------------------------------------\n"
        "La ………… valeur est NBNbWbWbNbWbNBNb de 24 / 30 milli\n"
        "grammes."
    )

    doc = blank_nlp(text)

    return doc


@mark.parametrize("ignore_excluded", [True, False])
@mark.parametrize("ignore_space_tokens", [True, False])
@mark.parametrize("attr", ["TEXT", "NORM"])
@mark.parametrize("full_doc", [True, False])
def test_create_span(
    doc2,
    ignore_excluded: bool,
    ignore_space_tokens: bool,
    attr: str,
    full_doc: bool,
):
    sent = list(doc2.sents)[1]
    doclike = doc2 if full_doc else sent

    matched_text = get_text(
        doclike,
        attr=attr,
        ignore_excluded=ignore_excluded,
        ignore_space_tokens=ignore_space_tokens,
    )
    clean_tokens = [
        t
        for t in doclike
        if not (
            (ignore_excluded and t.tag_ == "EXCLUDED")
            or (ignore_space_tokens and t.tag_ == "SPACE")
        )
    ]
    filtered_original = doc2[clean_tokens[0].i : clean_tokens[-1].i + 1].text
    for pattern, result, alignment_mode in [
        (r"4 / 3", "24 / 30", "expand"),
        (r"4 / 3", None, "strict"),
        (r"4 / 3", "/", "contract"),
        (r"24 / 30", "24 / 30", "expand"),
        (r"24 / 30", "24 / 30", "strict"),
        (r"24 / 30", "24 / 30", "contract"),
        (r"24 / 30 milli\s?gra", "24 / 30 milli\ngrammes", "expand"),
        (r"24 / 30 milli\s?gra", None, "strict"),
        (r"24 / 30 milli\s?gra", "24 / 30 milli\n", "contract"),
        (r" 24 / 30 ", "24 / 30", "expand"),
        (r" 24 / 30 ", None, "strict"),
        (r" 24 / 30 ", "24 / 30", "contract"),
        (matched_text, filtered_original, "expand"),
        (matched_text, filtered_original, "contract"),
        (matched_text, filtered_original, "strict"),
        ("(?=4 / 3)", "24", "expand"),
        ("(?=4 / 3)", None, "contract"),  # spacy behavior, but it's not ideal
        ("(?=4 / 3)", None, "strict"),
        ("(?=24)", "", "expand"),
        ("(?=24)", None, "contract"),  # spacy behavior, but it's not ideal
        ("(?=24)", None, "strict"),
    ]:
        match = re.search(pattern, matched_text)
        span = create_span(
            doclike,
            start_char=match.start(),
            end_char=match.end(),
            key="value",
            attr=attr,
            alignment_mode=alignment_mode,
            ignore_excluded=ignore_excluded,
            ignore_space_tokens=ignore_space_tokens,
        )
        assert (None if span is None else span.text) == result, (
            pattern,
            result,
            alignment_mode,
        )


def test_create_empty_span(blank_nlp):
    blank_nlp.add_pipe("eds.pollution")
    blank_nlp.add_pipe("eds.spaces")
    doc = blank_nlp("plan des addictions:\ntabac :0")

    span = create_span(
        doc[5:],
        0,
        0,
        "empty",
        attr="NORM",
        alignment_mode="expand",
        ignore_excluded=True,
        ignore_space_tokens=True,
    )
    assert span.start == 5 and span.end == 5


def test_empty_get_text(blank_nlp):
    blank_nlp.add_pipe("eds.pollution")
    blank_nlp.add_pipe("eds.spaces")
    doc = blank_nlp("==================================")
    clean = get_text(doc, attr="NORM", ignore_excluded=True, ignore_space_tokens=True)
    assert clean == ""



"""`eds.tnm` pipeline."""
from typing import Dict, List, Optional, Tuple, Union

from pydantic import ValidationError
from spacy.tokens import Doc, Span

from edsnlp.core import PipelineProtocol
from edsnlp.matchers.regex import RegexMatcher
from edsnlp.pipes.base import BaseNERComponent, SpanSetterArg
from edsnlp.utils.filter import filter_spans

from .model import TNM
from .patterns import tnm_pattern


class TNMMatcher(BaseNERComponent):
    """
    The `eds.tnm` component extracts [TNM](https://enwp.org/wiki/TNM_staging_system)
    mentions from clinical documents.

    Examples
    --------
    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.sentences())
    nlp.add_pipe(eds.tnm())

    text = "TNM: pTx N1 M1"

    doc = nlp(text)
    doc.ents
    # Out: (pTx N1 M1,)

    ent = doc.ents[0]
    ent._.tnm.dict()
    # {'modifier': 'p',
    #  'tumour': None,
    #  'tumour_specification': 'x',
    #  'node': '1',
    #  'node_specification': None,
    #  'metastasis': '1',
    #  'resection_completeness': None,
    #  'version': None,
    #  'version_year': None}
    ```

    Parameters
    ----------
    nlp : Optional[PipelineProtocol]
        The pipeline object
    name : str
        The name of the pipe
    pattern : Optional[Union[List[str], str]]
        The regex pattern to use for matching ADICAP codes
    attr : str
        Attribute to match on, eg `TEXT`, `NORM`, etc.
    label : str
        Label name to use for the `Span` object and the extension
    span_setter : SpanSetterArg
        How to set matches on the doc

    Authors and citation
    --------------------
    The TNM score is based on the development of S. Priou, B. Rance and
    E. Kempf ([@kempf:hal-03519085]).
    """

    # noinspection PyProtectedMember
    def __init__(
        self,
        nlp: Optional[PipelineProtocol],
        name: str = "tnm",
        *,
        pattern: Optional[Union[List[str], str]] = tnm_pattern,
        attr: str = "TEXT",
        label: str = "tnm",
        span_setter: SpanSetterArg = {"ents": True, "tnm": True},
    ):
        self.label = label

        super().__init__(nlp=nlp, name=name, span_setter=span_setter)

        if isinstance(pattern, str):
            pattern = [pattern]

        self.regex_matcher = RegexMatcher(attr=attr, alignment_mode="expand")
        self.regex_matcher.add(self.label, pattern)

    def set_extensions(self) -> None:
        """
        Set spaCy extensions
        """
        super().set_extensions()

        if not Span.has_extension(self.label):
            Span.set_extension(self.label, default=None)

    def process(self, doc: Doc) -> List[Span]:
        """
        Find TNM mentions in doc.

        Parameters
        ----------
        doc:
            spaCy Doc object

        Returns
        -------
        spans:
            list of tnm spans
        """

        spans = self.regex_matcher(
            doc,
            as_spans=True,
            return_groupdict=True,
        )

        filtered_spans = []
        for span, gd in spans:
            text = span.text
            clean = text.replace(" ", "").replace("\n", "")
            if (
                # we keep it if it's longer than 2 chars
                len(clean) > 2
                # or shorter but there is no space, and it starts w/ a lowercase letter
                # to avoid cases like "a  T" or "PT"
                or (not text[1:2].isspace() and text[0:1].islower())
            ):
                filtered_spans.append((span, gd))

        spans = filter_spans(filtered_spans)

        return spans

    def parse(self, spans: List[Tuple[Span, Dict[str, str]]]) -> List[Span]:
        """
        Parse dates using the groupdict returned by the matcher.

        Parameters
        ----------
        spans : List[Tuple[Span, Dict[str, str]]]
            List of tuples containing the spans and groupdict
            returned by the matcher.

        Returns
        -------
        List[Span]
            List of processed spans, with the date parsed.
        """

        for span, groupdict in spans:
            try:
                value = TNM.parse_obj(groupdict)
            except ValidationError:
                value = TNM.parse_obj({})

            span._.set(self.label, value)
            span.kb_id_ = value.norm()

        return [span for span, _ in spans]

    def __call__(self, doc: Doc) -> Doc:
        """
        Tags TNM mentions.

        Parameters
        ----------
        doc : Doc
            spaCy Doc object

        Returns
        -------
        doc : Doc
            spaCy Doc object, annotated for TNM
        """
        spans = self.process(doc)
        spans = self.parse(spans)
        self.set_spans(doc, spans)
        return doc



from io import StringIO
from typing import Dict, Optional, Union

import pandas as pd
from spacy.tokens import Doc, Span

from edsnlp.core import PipelineProtocol
from edsnlp.pipes.core.matcher.matcher import GenericMatcher
from edsnlp.pipes.misc.tables import patterns
from edsnlp.utils.filter import get_spans


class TablesMatcher(GenericMatcher):
    '''
    The `eds.tables` matcher detects tables in a documents.

    Examples
    --------
    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.normalizer())
    nlp.add_pipe(eds.tables())

    text = """
    SERVICE
    MEDECINE INTENSIVE –
    REANIMATION
    Réanimation / Surveillance Continue
    Médicale

    COMPTE RENDU D'HOSPITALISATION du 05/06/2020 au 10/06/2020
    Madame DUPONT Marie, née le 16/05/1900, âgée de 20 ans, a été hospitalisée en
    réanimation du 05/06/1920 au 10/06/1920 pour intoxication médicamenteuse volontaire.

    Examens complémentaires
    Hématologie
    Numération
    Leucocytes ¦x10*9/L ¦4.97 ¦4.09-11
    Hématies ¦x10*12/L¦4.68 ¦4.53-5.79
    Hémoglobine ¦g/dL ¦14.8 ¦13.4-16.7
    Hématocrite ¦% ¦44.2 ¦39.2-48.6
    VGM ¦fL ¦94.4 + ¦79.6-94
    TCMH ¦pg ¦31.6 ¦27.3-32.8
    CCMH ¦g/dL ¦33.5 ¦32.4-36.3
    Plaquettes ¦x10*9/L ¦191 ¦172-398
    VMP ¦fL ¦11.5 + ¦7.4-10.8

    Sur le plan neurologique : Devant la persistance d'une confusion à distance de
    l'intoxication au
    ...

    2/2Pat : <NOM> <Prenom>|F |<date> | <ipp> |Intitulé RCP
    """

    doc = nlp(text)

    # A table span
    table = doc.spans["tables"][0]

    # Leucocytes ¦x10*9/L ¦4.97 ¦4.09-11
    # Hématies ¦x10*12/L¦4.68 ¦4.53-5.79
    # Hémoglobine ¦g/dL ¦14.8 ¦13.4-16.7
    # Hématocrite ¦% ¦44.2 ¦39.2-48.6
    # VGM ¦fL ¦94.4 + ¦79.6-94
    # TCMH ¦pg ¦31.6 ¦27.3-32.8
    # CCMH ¦g/dL ¦33.5 ¦32.4-36.3
    # Plaquettes ¦x10*9/L ¦191 ¦172-398
    # VMP ¦fL ¦11.5 + ¦7.4-10.8

    # Convert span to Pandas table
    df = table._.to_pd_table()
    type(df)
    # Out: pandas.core.frame.DataFrame
    ```
    The pandas DataFrame:

    |      | 0           | 1        | 2      | 3         |
    | ---: | :---------- | :------- | :----- | :-------- |
    |    0 | Leucocytes  | x10*9/L  | 4.97   | 4.09-11   |
    |    1 | Hématies    | x10*12/L | 4.68   | 4.53-5.79 |
    |    2 | Hémoglobine | g/dL     | 14.8   | 13.4-16.7 |
    |    3 | Hématocrite | %        | 44.2   | 39.2-48.6 |
    |    4 | VGM         | fL       | 94.4 + | 79.6-94   |
    |    5 | TCMH        | pg       | 31.6   | 27.3-32.8 |
    |    6 | CCMH        | g/dL     | 33.5   | 32.4-36.3 |
    |    7 | Plaquettes  | x10*9/L  | 191    | 172-398   |
    |    8 | VMP         | fL       | 11.5 + | 7.4-10.8  |

    Extensions
    ----------
    The `eds.tables` pipeline declares the `span._.to_pd_table()` Span extension.
    This function returns a parsed pandas version of the table.

    Parameters
    ----------
    nlp : PipelineProtocol
        spaCy nlp pipeline to use for matching.
    name: str
        Name of the component.
    tables_pattern : Optional[Dict[str, str]]
        The regex pattern to identify tables.
        The key of dictionary should be `tables`
    attr : str
        spaCy's attribute to use:
        a string with the value "TEXT" or "NORM", or a dict with
        the key 'term_attr'. We can also add a key for each regex.
    ignore_excluded : bool
        Whether to skip excluded tokens.

    Authors and citation
    --------------------
    The `eds.tables` pipeline was developed by AP-HP's Data Science team.
    '''

    def __init__(
        self,
        nlp: PipelineProtocol,
        name: str = "tables",
        *,
        tables_pattern: Optional[Dict[str, str]] = None,
        sep_pattern: Optional[str] = None,
        attr: Union[Dict[str, str], str] = "TEXT",
        ignore_excluded: bool = True,
    ):
        if tables_pattern is None:
            self.tables_pattern = patterns.regex
        else:
            self.tables_pattern = tables_pattern

        if sep_pattern is None:
            self.sep = patterns.sep
        else:
            self.sep = sep_pattern

        super().__init__(
            nlp=nlp,
            name=name,
            terms=None,
            regex=self.tables_pattern,
            attr=attr,
            ignore_excluded=ignore_excluded,
        )

        if not Span.has_extension("to_pd_table"):
            Span.set_extension("to_pd_table", method=self.to_pd_table)

        self.set_extensions()

    def __call__(self, doc: Doc) -> Doc:
        """Find spans that contain tables

        Parameters
        ----------
        doc : Doc

        Returns
        -------
        Doc
        """
        matches = self.process(doc)
        tables = get_spans(matches, "tables")
        # parsed = self.parse(tables=tables)

        doc.spans["tables"] = tables

        return doc

    def to_pd_table(self, span) -> pd.DataFrame:
        table_str_io = StringIO(span.text)
        parsed = pd.read_csv(
            table_str_io,
            sep=self.sep,
            engine="python",
            header=None,
            on_bad_lines="skip",
        )
        return parsed
